"""Lateral operators."""

# Standard library
from typing import Literal

# Third-party
import numpy as np
import xarray as xr


def compute_weights(
    window_size: int,
    window_type: Literal["exp", "const"],
    window_shape: Literal["disk", "square"],
) -> xr.DataArray:
    """Compute weights for a convolution kernel.

    Parameters
    ----------
    window_size : int
        Number of grid cells the kernel. Must be an odd number.
    window_type : {"exp", "const"}
        The window type.

        Supported types:
         - exp: Exponential drop from the center of the window.
         - const: All weights are set to 1.

    window_shape : {"disk", "square"}
        The window shape.

        Supported shapes:
         - disk: The kernel is confined to the circle inscribing the window.
         - square: The kernel is defined on the entire window.

    Raises
    ------
    ValueError
        if `window_size` is not an odd number.

    Returns
    -------
    xarray.DataArray
        Weights corresponding to the `window_type` within the given `window_shape`
        and zero outside of the shape.

    """
    n = window_size

    if n % 2 != 1:
        raise ValueError("window_size must be an odd number.")

    radius = n // 2
    yy, xx = np.mgrid[:n, :n] - radius
    dist = np.sqrt(xx**2 + yy**2)

    if window_type == "exp":
        kernel = np.exp(-dist)
    else:
        kernel = np.ones((n, n))

    weights = xr.DataArray(kernel, dims=["win_x", "win_y"])

    if window_shape == "disk":
        return weights.where(dist <= radius, 0.0)

    return weights


def compute_cond_mask(
    windows: xr.DataArray, weights: xr.DataArray, frac_val: float, nx: int, ny: int
) -> xr.DataArray:
    """Compute the conditional mask for which the convolution results are valid.

    The function computes the ratio of undefined values to the total number values
    in the convolution window. The computational domain and the window shape are both
    taken into account. The mask is true for locations where the ratio is above the
    given threshold.

    Parameters
    ----------
    windows : xarray.DataArray
        Values of the constructed windows for the application of the convolution.
    weights : xarray.DataArray
        Weights of the convolution kernel.
    frac_val : float
        Threshold on the ratio of undefined values.
    nx : int
        Size of the field in the x dimension.
    ny : int
        Size of the field in the y dimension.

    Returns
    -------
    xarray.DataArray
        Boolean mask indicating where the undefined value ratio in the convolution
        kernel is within the acceptable threshold.

    """
    # For each window, `x` is the coordinate of the center of the window in the parent
    # field. `win_x` is the local coordinate of the current element in the window.
    # `loc_x` represents the location of the window element in the coordinates of the
    # parent field. This is needed to filter out nans that are introduced by the
    # padding of the field by the rolling method. Only the nans that are present within
    # the bound of the field should count towards the `frac_val` threshold.

    mask = weights > 0

    loc_x = windows.x + windows.win_x - windows.sizes["win_x"] // 2
    in_bnds_x = np.logical_and(0 <= loc_x, loc_x < nx)

    loc_y = windows.y + windows.win_y - windows.sizes["win_y"] // 2
    in_bnds_y = np.logical_and(0 <= loc_y, loc_y < ny)

    undef = windows.isnull()
    frac_undef = undef.where(mask).where(in_bnds_x).where(in_bnds_y).mean(weights.dims)
    return frac_undef <= frac_val


def fill_undef(field: xr.DataArray, radius: int, frac_val: float) -> xr.DataArray:
    """Fill undefined values.

    Replace undefined values with a value derived from neighbourhood.
    The values are obtained by applying a Gaussian filter to the field
    around the undefined elements.

    Parameters
    ----------
    field : xarray.DataArray
        Input field.
    radius : int
        Radius of the convolution window.
    frac_val : float
        Threshold of acceptable ratio of undefined values in window.
        If the ratio of undefined to total elements in the window is
        below the threshold then the output value remains undefined.

    Returns
    -------
    xarray.DataArray
        Input field with undefined values replaced by values derived from neighbourhood.

    """
    n = 2 * radius + 1
    weights = compute_weights(n, "exp", "disk")

    # select undefined grid elements
    undef = field.isnull()
    idx = {
        dim: xr.DataArray(idx, dims="undef")
        for dim, idx in zip(undef.dims, np.nonzero(undef.values))
    }
    xy_coords = {dim: range(field.sizes[dim]) for dim in "xy"}

    # construct rolling windows
    dims = {"x": "win_x", "y": "win_y"}
    windows = (
        field.assign_coords(xy_coords)
        .rolling({"x": n, "y": n}, center=True)
        .construct(dims)
        .isel(idx)
    )

    # compute conditional mask
    s = field.sizes
    cond = compute_cond_mask(windows, weights, frac_val, s["x"], s["y"])

    # compute weighted mean skipping undefined values
    smoothed = windows.weighted(weights).mean(dims.values())

    # replace undefined values in input field
    result = field.copy()
    result[idx] = smoothed.where(cond)
    return result


def disk_avg(field: xr.DataArray, radius: int) -> xr.DataArray:
    """Compute disk average.

    Parameters
    ----------
    field : xarray.DataArray
        Input field.
    radius : int
        Radius of the convolution window.

    Returns
    -------
    xarray.DataArray

    """
    n = 2 * radius + 1
    weights = compute_weights(n, "const", "disk")

    # construct rolling windows
    dims = {"x": "win_x", "y": "win_y"}
    windows = field.rolling({"x": n, "y": n}, center=True).construct(dims)

    # compute weighted mean skipping undefined values
    skipna = field.isnull().any().item()
    smoothed = windows.weighted(weights).mean(dims.values(), skipna=skipna)

    return smoothed.where(~field.isnull())



"""Definition of the thetav operator."""

# Third-party
import xarray as xr

# Local
from .. import metadata
from .. import physical_constants as pc


def fthetav(p: xr.DataArray, t: xr.DataArray, qv: xr.DataArray) -> xr.DataArray:
    """Virtual potential temperature in K.

    Parameters
    ----------
    p : xarray.DataArray
        pressure in Pa
    t : xarray.DataArray
        air temperature in K
    qv : xarray.DataArray
        specific humidity (dimensionless)

    Returns
    -------
    xarray.DataArray
        virtual potential temperature in K

    """
    # Reference surface pressure for computation of potential temperature
    p0 = 1.0e5

    return xr.DataArray(
        data=(p0 / p) ** pc.rdocp * t * (1.0 + (pc.rvd_o * qv / (1.0 - qv))),
        attrs=metadata.override(t.metadata, shortName="THETA_V"),
    )



"""functionality for tasking and parallel computing."""

# Third-party
import dask

# First-party
import meteodatalab.config


def delayed(fn):
    return (
        dask.delayed(fn, pure=True)
        if meteodatalab.config.get("enable_dask", False)
        else fn
    )


def compute(*delayed_objs):
    return (
        dask.compute(*delayed_objs)
        if meteodatalab.config.get("enable_dask", False)
        else delayed_objs
    )



"""Differential operators on a regular grid."""

# Standard library
import dataclasses as dc

# Third-party
import numpy as np
import xarray as xr


@dc.dataclass
class _Slicer:
    ndim: int
    axis: int

    def __getitem__(self, ind):
        slices = [slice(None)] * self.ndim
        slices[self.axis] = ind
        return tuple(slices)


def _dxy(field: xr.DataArray, axis: int) -> np.ndarray:
    s = _Slicer(ndim=len(field.dims), axis=axis)
    df_dxy = np.empty_like(field)
    df_dxy[s[1:-1]] = 0.5 * (field[s[2:]] - field[s[:-2]])
    df_dxy[s[0]] = np.nan
    df_dxy[s[-1]] = np.nan
    return df_dxy


def dx(field: xr.DataArray) -> xr.DataArray:
    """Compute central finite difference along x-axis.

    The x-axis is assumed to be the last dimension of the array.
    The coordinates spacing is assumed to be of unit length.
    The field is assumed to be located at the mass points.
    The first and last elements in the x direction are undefined.
    The returned array is the same shape as the input array.

    Parameters
    ----------
    field: xr.DataArray
        The field for which to compute the finite difference along the x-axis.

    Returns
    -------
    xr.DataArray
        The finite difference of the field along the x-axis.

    """
    return field.copy(data=_dxy(field, axis=-1))


def dy(field: xr.DataArray) -> xr.DataArray:
    """Compute central finite difference along y-axis.

    The y-axis is assumed to be the last dimension of the array.
    The coordinates spacing is assumed to be of unit length.
    The field is assumed to be located at the mass points.
    The first and last elements in the y direction are undefined.
    The returned array is the same shape as the input array.

    Parameters
    ----------
    field: xr.DataArray
        The field for which to compute the finite difference along the y-axis.

    Returns
    -------
    xr.DataArray
        The finite difference of the field along the y-axis.

    """
    return field.copy(data=_dxy(field, axis=-2))


def dz(field: xr.DataArray) -> xr.DataArray:
    """Compute central finite difference along z-axis.

    The z-axis is assumed to be the third to last dimension of the array.
    The coordinates spacing is assumed to be of unit length.
    The field is assumed to be located at the mass points.
    The first and last elements in the z direction are computed using a first order
    scheme.
    The returned array is the same shape as the input array.

    Parameters
    ----------
    field: xr.DataArray
        The field for which to compute the finite difference along the z-axis.

    Returns
    -------
    xr.DataArray
        The finite difference of the field along the z-axis.

    """
    return field.copy(data=np.gradient(field, axis=-3))


def dz_staggered(field: xr.DataArray) -> xr.DataArray:
    """Compute central finite difference along z-axis on staggered field.

    The z-axis is assumed to be the third to last dimension of the array.
    The coordinates spacing is assumed to be of unit length.
    The field is assumed to be located at staggered positions in the z direction.
    The returned array is located on the mass points.

    Parameters
    ----------
    field: xr.DataArray
        The field for which to compute the finite difference along the z-axis.

    Returns
    -------
    xr.DataArray
        The finite difference of the field along the z-axis.

    """
    return field.diff(dim="z", label="lower")



"""Atmospheric thermodynamic functions."""

# Third-party
import numpy as np

# First-party
from meteodatalab import physical_constants as pc


def pv_sw(t):
    """Pressure of water vapor at equilibrium over liquid water.

    Parameters
    ----------
    t : xarray.DataArray
        temperature (in Kelvin)

    Returns
    -------
    xarray.DataArray
        pressure of water vapor in Pascal

    """
    return pc.b1 * np.exp(pc.b2w * (t - pc.b3) / (t - pc.b4w))


def pv_si(t):
    """Pressure of water vapor at equilibrium over ice.

    Parameters
    ----------
    t : xarray.DataArray
        temperature (in Kelvin)

    Returns
    -------
    xarray.DataArray
        pressure of water vapor in Pascal

    """
    return pc.b1 * np.exp(pc.b2i * (t - pc.b3) / (t - pc.b4i))


def pv_sm(t):
    """Pressure of water vapor at equilibrium over mixed phase.

    Parameters
    ----------
    t : xarray.DataArray
        temperature (in Kelvin)

    Returns
    -------
    xarray.DataArray
        pressure of water vapor in Pascal

    """
    tice_only = pc.b3 - 23
    dtice = pc.b3 - tice_only
    alpha = ((t - tice_only) / dtice) ** 2
    water = pv_sw(t)
    ice = pv_si(t)
    mix = alpha * water + (1.0 - alpha) * ice
    pv_sm = water.where(t > pc.b3, ice.where(t < tice_only, mix))
    return pv_sm


def qv_pvp(pv, p):
    """Specific water vapor content (from perfect gas law and approximating q~w).

    Parameters
    ----------
    pv : xarray.DataArray
        pressure of water vapor
    p : xarray.DataArray
        pressure

    Returns
    -------
    xarray.DataArray
        specific water vapor (dimensionless)

    """
    return pc.rdv * pv / np.maximum((p - pc.o_rdv * pv), 1.0)


def pv_qp(qv, p):
    """Partial pressure of water vapor (from perfect gas law and approximating q~w).

    Parameters
    ----------
    qv : xarray.DataArray
        water vapor mixing ratio
    p : xarray.DataArray
        pressure

    Returns
    -------
    xarray.DataArray
        partial pressure of water vapor (same unit as p)

    """
    return qv * p / (pc.rdv + pc.o_rdv * qv)



"""algorithm for BRN operator."""

# Third-party
import numpy as np
import xarray as xr

# Local
from .. import metadata
from .. import physical_constants as pc
from .destagger import destagger
from .thetav import fthetav


def fbrn(
    p: xr.DataArray,
    t: xr.DataArray,
    qv: xr.DataArray,
    u: xr.DataArray,
    v: xr.DataArray,
    hhl: xr.DataArray,
    hsurf: xr.DataArray,
) -> xr.DataArray:
    """Bulk Richardson Number (BRN).

    Parameters
    ----------
    p : xarray.DataArray
        pressure in Pa
    t : xarray.DataArray
        air temperature in K
    qv : xarray.DataArray
        specific humidity (dimensionless)
    u : xarray.DataArray
        the x component of the wind velocity in m/s
    v : xarray.DataArray
        the y component of the wind velocity in m/s
    hhl : xarray.DataArray
        Heights of the interfaces between vertical layers in m amsl
    hsurf : xarray.DataArray
        earth surface height in m amsl


    Returns
    -------
    xarray.DataArray
        Bulk Richardson Number (dimensionless)

    """
    nlevels = p.sizes["z"]

    thetav = fthetav(p, t, qv)
    thetav_sum = (
        thetav.isel(z=slice(None, None, -1))
        .cumsum(dim="z")
        .isel(z=slice(None, None, -1))
    )

    nlevels_xr = xr.DataArray(data=np.arange(nlevels, 0, -1), dims=["z"])
    u_ = destagger(u, "x")
    v_ = destagger(v, "y")
    hfl = destagger(hhl, "z")

    brn = (
        pc.g
        * (hfl - hsurf)
        * (thetav - thetav.isel(z=nlevels - 1))
        * nlevels_xr
        / (thetav_sum * (u_**2 + v_**2))
    )

    return xr.DataArray(
        data=brn,
        attrs=metadata.override(p.metadata, shortName="BRN"),
    )



"""Geospatial logic."""

# Standard library
import dataclasses as dc
import typing
from collections.abc import Mapping

# Third-party
import numpy as np
import xarray as xr

# Local
from .. import metadata
from ..grib_decoder import set_code_flag
from .support_operators import get_grid_coords


@dc.dataclass
class RotLatLonGrid:
    """Class representing a rotated lat lon grid.

    Attributes
    ----------
    rlon : xr.DataArray
        longitude values in degrees of the grid points in the rotated lat lon CRS.
    rlat : xr.DataArray
        latitude values in degrees of the grid points in the rotated lat lon CRS.
    north_pole_lon : float
        longitude of the rotated north pole in degrees.
    north_pole_lat : float
        latitude of the rotated north pole in degrees.

    """

    # all units in degrees
    rlon: xr.DataArray
    rlat: xr.DataArray
    north_pole_lon: float
    north_pole_lat: float


def _check_requirements(geo: Mapping[str, typing.Any]) -> None:
    requirements = {
        "angleOfRotationInDegrees": 0.0,
        "gridType": "rotated_ll",
        "iScansNegatively": 0,
        "jScansPositively": 1,
    }
    errors = {key: geo[key] for key in requirements if requirements[key] != geo[key]}
    if errors:
        msg = f"Unsupported values for keys: {errors}"
        raise ValueError(msg)


def get_grid(geo: Mapping[str, typing.Any]) -> RotLatLonGrid:
    """Get grid parameters for a given field.

    Only fields defined on regular grids in rotlatlon coordinates,
    without rotation nor flipped axes are supported.

    Parameters
    ----------
    geo : Mapping[str, Any]
        Grib keys related to the geography of the field.

    Raises
    ------
    ValueError
        if the field does not fulfill the conditions above.

    Returns
    -------
    RotLatLonGrid
        object representing the rotated lat lon grid.

    """
    _check_requirements(geo)

    ni = geo["Ni"]
    x0 = geo["longitudeOfFirstGridPointInDegrees"]
    dx = geo["iDirectionIncrementInDegrees"]
    rlon = get_grid_coords(ni, x0, dx, "x")

    nj = geo["Nj"]
    y0 = geo["latitudeOfFirstGridPointInDegrees"]
    dy = geo["jDirectionIncrementInDegrees"]
    rlat = get_grid_coords(nj, y0, dy, "y")

    lon_np = (geo["longitudeOfSouthernPoleInDegrees"] - 180) % 360
    lat_np = -1 * geo["latitudeOfSouthernPoleInDegrees"]

    return RotLatLonGrid(rlon, rlat, lon_np, lat_np)


def rot2geolatlon(grid: RotLatLonGrid) -> tuple[xr.DataArray, xr.DataArray]:
    """Compute geographical lat lon values for a rotated lat lon grid.

    Parameters
    ----------
    grid : RotLatLonGrid
        object representing the rotated lat lon grid.

    Returns
    -------
    tuple[xarray.DataArray, xarray.DataArray]
        tuple of longitudes and latitudes of the grid points in the
        geographical lat lon CRS.

    """
    deg2rad = np.pi / 180
    rad2deg = 180 / np.pi
    sin_np_lat = np.sin(deg2rad * grid.north_pole_lat)
    cos_np_lat = np.cos(deg2rad * grid.north_pole_lat)

    sin_np_lon = np.sin(deg2rad * grid.north_pole_lon)
    cos_np_lon = np.cos(deg2rad * grid.north_pole_lon)

    # Compute new coordinates
    # ... normalize input coordinates
    norm_lon = deg2rad * grid.rlon.where(grid.rlon < 180, grid.rlon - 360)

    # ... cache trigonometric operations
    sin_lat = np.sin(deg2rad * grid.rlat)
    cos_lat = np.cos(deg2rad * grid.rlat)
    sin_lon = np.sin(norm_lon)
    cos_lon = np.cos(norm_lon)
    # ... compute latitude
    arg1 = cos_np_lat * cos_lat * cos_lon + sin_np_lat * sin_lat
    lat = rad2deg * np.arcsin(arg1)
    # ... compute longitude
    arg2 = (
        sin_np_lon * (-sin_np_lat * cos_lon * cos_lat + cos_np_lat * sin_lat)
        - cos_np_lon * sin_lon * cos_lat
    )
    arg3 = (
        cos_np_lon * (-sin_np_lat * cos_lon * cos_lat + cos_np_lat * sin_lat)
        + sin_np_lon * sin_lon * cos_lat
    )
    # BUG: changes sign when arg2 is negative and less than threshold
    arg4 = xr.where(np.abs(arg3) < 1e-20, 1e-20, arg3)
    lon = rad2deg * np.arctan2(arg2, arg4) % 360

    return lon, lat


def geolatlon2swiss(
    lon: xr.DataArray, lat: xr.DataArray
) -> tuple[xr.DataArray, xr.DataArray]:
    """Convert from geolatlon to swiss coordinates.

    Parameters
    ----------
    lon : xarray.DataArray
        longitude coordinates in the geolatlon CRS.
    lat : xarray.DataArray
        latitude coordinates in the geolatlon CRS.

    Returns
    -------
    tuple[xarray.DataArray, xarray.DataArray]
        x and y coordinates in the Swiss LV03 coordinate system.

    Notes
    -----
    Approximate formula published by swisstopo, precision in the order of 1 meter

    """
    norm_lat = ((lat * 3.6) - 169.02866) / 10
    lon = lon.where(lon < 180, lon - 360)
    norm_lon = ((lon * 3.6) - 26.7825) / 10
    y = (
        200147.07
        + 3745.25 * norm_lon * norm_lon
        + norm_lat
        * (
            308807.95
            + 76.63 * norm_lat
            - 194.56 * norm_lon * norm_lon
            + 119.79 * norm_lat * norm_lat
        )
    )
    x = 600072.37 + norm_lon * (
        211455.93
        - 10938.51 * norm_lat
        - 0.36 * norm_lat * norm_lat
        - 44.54 * norm_lon * norm_lon
    )
    return x, y


def vref_rot2geolatlon(
    u: xr.DataArray, v: xr.DataArray
) -> tuple[xr.DataArray, xr.DataArray]:
    """Apply coordinate rotation to vector field.

    When converting from rotated lat lon to geo lat lon, the
    orientation of the grid changes and vector fields for which
    the components are expressed in the grid unit vectors need
    to be realigned.

    Note that this function does not perform any regridding.

    Parameters
    ----------
    u : xarray.DataArray
        x component of the vector field w.r.t. a rotated lat lon grid.
    v : xarray.DataArray
        y component of the vector field w.r.t. a rotated lat lon grid.

    Raises
    ------
    ValueError
        if either field is on a staggered grid or other than rotated lat lon.

    Returns
    -------
    tuple[xarray.DataArray, xarray.DataArray]
        x and y components of the vector field w.r.t. the geo lat lon coords.

    """
    if metadata.is_staggered_horizontal(u) or metadata.is_staggered_horizontal(v):
        raise ValueError("The vector fields must be destaggered.")
    if (
        u.metadata.get("gridDefinitionTemplateNumber") != 1
        or v.metadata.get("gridDefinitionTemplateNumber") != 1
    ):
        raise ValueError("The vector fields must be defined on a rotated lat lon grid.")

    grid = get_grid(u.geography)
    lon, lat = rot2geolatlon(grid)
    u_g, v_g = _vref_rot2geolatlon(u, v, lon, lat, grid)

    # bit 5 left unset since u/v relative to grid in x (i) and y (j) directions,
    # not defined grid; bits 3 and 4 set as i, j direction increments given
    resolution_components_flags = set_code_flag([3, 4])

    return (
        xr.DataArray(
            u_g,
            attrs=metadata.override(
                u.metadata, resolutionAndComponentFlags=resolution_components_flags
            ),
        ),
        xr.DataArray(
            v_g,
            attrs=metadata.override(
                v.metadata, resolutionAndComponentFlags=resolution_components_flags
            ),
        ),
    )


def _vref_rot2geolatlon(
    u: xr.DataArray,
    v: xr.DataArray,
    lon: xr.DataArray,
    lat: xr.DataArray,
    grid: RotLatLonGrid,
) -> tuple[xr.DataArray, xr.DataArray]:
    deg2rad = np.pi / 180
    sin_np = np.sin(deg2rad * grid.north_pole_lat)
    cos_np = np.cos(deg2rad * grid.north_pole_lat)

    norm_lat = lat * deg2rad
    norm_dlon = (grid.north_pole_lon - lon) * deg2rad
    arg1 = cos_np * np.sin(norm_dlon)
    arg2 = sin_np * np.cos(norm_lat) - cos_np * np.sin(norm_lat) * np.cos(norm_dlon)
    norm = 1.0 / np.sqrt(arg1**2 + arg2**2)

    u_out = u * arg2 * norm + v * arg1 * norm
    v_out = -u * arg1 * norm + v * arg2 * norm

    return u_out, v_out



"""Radiation related operators."""

# Third-party
import xarray as xr

# Local
from .. import metadata
from .. import physical_constants as pc


def compute_athd_s(athb_s: xr.DataArray, tsurf: xr.DataArray) -> xr.DataArray:
    """Compute incoming longwave radiation at surface level.

    Parameters
    ----------
    athb_s : xarray.DataArray
        Net long-wave radiation flux at surface [W m-2]
    tsurf : xarray.DataArray
        Temperature at surface [K]

    Returns
    -------
    xarray.DataArray
        Average downward longwave radiation at the surface [W m-2]

    """
    return xr.DataArray(
        data=athb_s / pc.emissivity_surface + pc.boltzman_cst * tsurf**4,
        attrs=metadata.override(athb_s.metadata, shortName="ATHD_S"),
    )


def compute_swdown(diffuse: xr.DataArray, direct: xr.DataArray) -> xr.DataArray:
    """Compute downward shortwave radiation.

    Parameters
    ----------
    diffuse : xarray.DataArray
        incoming diffuse shortwave radiation.
    direct : xarray.DataArray
        incoming direct shortwave radiation.

    Returns
    -------
    xarray.DataArray
        downward shortwave radiation at surface level.

    """
    return xr.DataArray(
        data=(diffuse + direct).clip(min=0),
        attrs=metadata.override(diffuse.metadata, shortName="ASOD_S"),
    )



"""Regridding operator."""

# Standard library
import dataclasses as dc
import typing
import warnings
from typing import Literal

# Third-party
import numpy as np
import xarray as xr
from numpy.typing import ArrayLike, NDArray

try:
    # Third-party
    from pyproj import Transformer
    from rasterio import transform, warp
    from rasterio.crs import CRS
    from scipy.spatial import Delaunay  # type: ignore
except ImportError:
    raise ImportError("The regrid operator requires extra dependencies.")

# Local
from .. import icon_grid, metadata
from ..grib_decoder import set_code_flag

Resampling: typing.TypeAlias = warp.Resampling

# For more information: check https://epsg.io/<id>
CRS_ALIASES = {
    "geolatlon": "epsg:4326",  # WGS84
    "swiss": "epsg:21781",  # Swiss CH1903 / LV03
    "swiss03": "epsg:21781",  # Swiss CH1903 / LV03
    "swiss95": "epsg:2056",  # Swiss CH1903+ / LV95
    "boaga-west": "epsg:3003",  # Monte Mario / Italy zone 1
    "boaga-east": "epsg:3004",  # Monte Mario / Italy zone 2
    "utm32n": "epsg:32632",  # Universal Transverse Mercator zone 32N
}


def _get_crs(geo):
    if geo["gridType"] != "rotated_ll":
        raise NotImplementedError("Unsupported grid type")

    lon = geo["longitudeOfSouthernPoleInDegrees"]
    lat = -1 * geo["latitudeOfSouthernPoleInDegrees"]

    return CRS.from_string(
        f"+proj=ob_tran +o_proj=longlat +o_lat_p={lat} +lon_0={lon} +datum=WGS84"
    )


def _normalise(angle: float) -> float:
    return np.fmod(angle + 180, 360) - 180


@dc.dataclass
class RegularGrid:
    """Class defining a regular grid.

    Attributes
    ----------
    crs : CRS
        Coordinate reference system.
    nx : int
        Number of grid points in the x direction.
    ny : int
        Number of grid points in the y direction.
    xmin : float
        Coordinate of the first grid point in the x direction.
    xmax : float
        Coordinate of the last grid point in the x direction.
    ymin : float
        Coordinate of the first grid point in the y direction.
    ymax : float
        Coordinate of the last grid point in the y direction.

    """

    crs: CRS
    nx: int
    ny: int
    xmin: float
    xmax: float
    ymin: float
    ymax: float

    @classmethod
    def from_field(cls, field: xr.DataArray):
        """Extract grid parameters from grib metadata.

        Parameters
        ----------
        field : xarray.DataArray
            field containing the relevant metadata.

        """
        geo = field.geography
        obj = cls(
            crs=_get_crs(geo),
            nx=geo["Ni"],
            ny=geo["Nj"],
            xmin=_normalise(geo["longitudeOfFirstGridPointInDegrees"]),
            xmax=_normalise(geo["longitudeOfLastGridPointInDegrees"]),
            ymin=geo["latitudeOfFirstGridPointInDegrees"],
            ymax=geo["latitudeOfLastGridPointInDegrees"],
        )
        if abs(obj.dx - geo["iDirectionIncrementInDegrees"]) > 1e-5:
            raise ValueError("Inconsistent grid parameters")
        if abs(obj.dy - geo["jDirectionIncrementInDegrees"]) > 1e-5:
            raise ValueError("Inconsistent grid parameters")
        return obj

    @classmethod
    def parse_regrid_operator(cls, op: str):
        """Parse fieldextra out_regrid_target string.

        Parameters
        ----------
        op : str
            fieldextra out_regrid_target definition
            i.e. crs,xmin,ymin,xmay,ymax,dx,dy.

        """
        crs_str, *grid_params = op.split(",")
        crs = CRS.from_string(CRS_ALIASES[crs_str])
        xmin, ymin, xmax, ymax, dx, dy = map(float, grid_params)
        if abs(dx) < 1e-10 or abs(dy) < 1e-10:
            raise ValueError("Inconsistent regrid parameters")
        nx = (xmax - xmin) / dx + 1
        ny = (ymax - ymin) / dy + 1
        if abs(nx - round(nx)) > 1e-10 or abs(ny - round(ny)) > 1e-10:
            raise ValueError("Inconsistent regrid parameters")
        return cls(crs, int(round(nx)), int(round(ny)), xmin, xmax, ymin, ymax)

    def to_crs(self, crs: str, **kwargs):
        """Return a new grid in the given coordinate reference system.

        Parameters
        ----------
        crs : str
            The coordinate reference system in which the output grid should be defined.

        Returns
        -------
        RegularGrid or subtype
            Output grid in the given coordinate reference system.

        """
        dst_crs = CRS.from_string(crs)
        tx, width, height = typing.cast(
            tuple[transform.Affine, int, int],
            warp.calculate_default_transform(
                src_crs=self.crs,
                dst_crs=dst_crs,
                width=self.nx,
                height=self.ny,
                left=self.xmin,
                bottom=self.ymin,
                right=self.xmax,
                top=self.ymax,
                **kwargs,
            ),
        )
        a = transform.AffineTransformer(tx)
        xmin, ymax = a.xy(-0.5, -0.5)
        xmax, ymin = a.xy(height + 0.5, width + 0.5)  # row, col
        return type(self)(dst_crs, width, height, xmin, xmax, ymin, ymax)

    @property
    def x(self) -> np.ndarray:
        return np.arange(self.nx) * self.dx + self.xmin

    @property
    def y(self) -> np.ndarray:
        return np.arange(self.ny) * self.dy + self.ymin

    @property
    def dx(self) -> float:
        return (self.xmax - self.xmin) / (self.nx - 1)

    @property
    def dy(self) -> float:
        return (self.ymax - self.ymin) / (self.ny - 1)

    @property
    def transform(self) -> transform.Affine:
        return transform.from_origin(
            west=self.xmin - self.dx / 2,
            north=self.ymax + self.dy / 2,
            xsize=self.dx,
            ysize=self.dy,
        )


def _udeg(value: float) -> int:
    return int(round(value * 1e6))


def _lon_from_utm_zone(zone: int) -> int:
    if zone is None or zone < 1 or zone > 60:
        raise ValueError(f"Invalid value for UTM zone: {zone}.")

    # UTM zone are 6 degrees wide numbered from 1 starting at -180 to -174 degrees.
    offset = zone * 6
    return offset - 183


def _grib_utm_m(value: float | int) -> int:
    # For UTM units, GRIB2 uses 10-2m whereas CRS uses m.
    return int(round(100 * value))


def _get_metadata(grid: RegularGrid):
    if grid.crs.to_epsg() == 4326:
        # geolatlon
        # https://codes.ecmwf.int/grib/format/grib2/ctables/3/4/
        scanning_mode = set_code_flag([2])  # positive y
        # i, j direction increments given
        # bit 5 left unset since the target grid is geolatlon
        # both values are equivalent in this case
        resolution_components_flags = set_code_flag([3, 4])
        return {
            "numberOfDataPoints": grid.nx * grid.ny,
            "sourceOfGridDefinition": 0,  # defined by template number
            "numberOfOctectsForNumberOfPoints": 0,
            "interpretationOfNumberOfPoints": 0,
            "gridDefinitionTemplateNumber": 0,  # latlon
            "shapeOfTheEarth": 5,  # WGS 84
            "Ni": grid.nx,
            "Nj": grid.ny,
            "latitudeOfFirstGridPoint": _udeg(grid.ymin),
            "longitudeOfFirstGridPoint": _udeg(grid.xmin),
            "resolutionAndComponentFlags": resolution_components_flags,
            "latitudeOfLastGridPoint": _udeg(grid.ymax),
            "longitudeOfLastGridPoint": _udeg(grid.xmax),
            "iDirectionIncrement": _udeg(grid.dx),
            "jDirectionIncrement": _udeg(grid.dy),
            "scanningMode": scanning_mode,
        }
    elif grid.crs.get("proj") == "utm" and not grid.crs.get("south"):
        # Transverse Mercator in northern hemisphere.
        # https://codes.ecmwf.int/grib/format/grib2/ctables/3/4/
        scanning_mode = set_code_flag([2])  # positive y
        # i, j direction increments given relative to defined grid.
        resolution_components_flags = set_code_flag([3, 4, 5])
        return {
            "numberOfDataPoints": grid.nx * grid.ny,
            "sourceOfGridDefinition": 0,  # defined by template number
            "numberOfOctectsForNumberOfPoints": 0,
            "interpretationOfNumberOfPoints": 0,
            "gridDefinitionTemplateNumber": 12,  # UTM
            "shapeOfTheEarth": 5,  # WGS 84
            "Ni": grid.nx,
            "Nj": grid.ny,
            "falseEasting": _grib_utm_m(500000),
            "falseNorthing": 0,  # Northern hemisphere
            "scaleFactorAtReferencePoint": 0.9996,
            "longitudeOfReferencePoint": _lon_from_utm_zone(grid.crs.get("zone")),
            "Di": _grib_utm_m(grid.dx),
            "Dj": _grib_utm_m(grid.dy),
            "X1": _grib_utm_m(grid.xmin),
            "X2": _grib_utm_m(grid.xmax),
            "Y1": _grib_utm_m(grid.ymin),
            "Y2": _grib_utm_m(grid.ymax),
            "resolutionAndComponentFlags": resolution_components_flags,
            "scanningMode": scanning_mode,
        }
    elif grid.crs.get("proj") == "ob_tran":
        # rotlatlon
        # https://codes.ecmwf.int/grib/format/grib2/ctables/3/4/
        scanning_mode = set_code_flag([2])  # positive y
        # i, j direction increments given
        resolution_components_flags = set_code_flag([3, 4])
        return {
            "numberOfDataPoints": grid.nx * grid.ny,
            "sourceOfGridDefinition": 0,  # defined by template number
            "numberOfOctectsForNumberOfPoints": 0,
            "interpretationOfNumberOfPoints": 0,
            "gridDefinitionTemplateNumber": 1,  # rotlatlon
            "shapeOfTheEarth": 5,  # WGS 84
            "Ni": grid.nx,
            "Nj": grid.ny,
            "latitudeOfFirstGridPoint": _udeg(grid.ymin),
            "longitudeOfFirstGridPoint": _udeg(grid.xmin),
            "resolutionAndComponentFlags": resolution_components_flags,
            "latitudeOfLastGridPoint": _udeg(grid.ymax),
            "longitudeOfLastGridPoint": _udeg(grid.xmax),
            "iDirectionIncrement": _udeg(grid.dx),
            "jDirectionIncrement": _udeg(grid.dy),
            "scanningMode": scanning_mode,
            "latitudeOfSouthernPole": _udeg(-1 * grid.crs.get("o_lat_p")),
            "longitudeOfSouthernPole": _udeg(grid.crs.get("lon_0")),
            "angleOfRotation": 0.0,
        }
    else:
        # Rather than attempt to populate the GRIB definition in the general case, we
        # just set the sourceOfGridDefinition to 255. Ideally we would also set
        # gridDefinitionTemplateNumber to 65535, but eccodes attempts and fails to find
        # a template by this number.
        return {
            "sourceOfGridDefinition": 255,  # grid definition does not apply
            "numberOfDataPoints": grid.nx * grid.ny,
            "numberOfOctectsForNumberOfPoints": 0,
        }


def regrid(
    field: xr.DataArray,
    dst: RegularGrid,
    resampling: Resampling,
    src: RegularGrid | None = None,
):
    """Regrid a field.

    Parameters
    ----------
    field : xarray.DataArray
        Input field defined on a regular grid in rotated latlon coordinates.
    dst : RegularGrid
        Destination grid onto which to project the field.
    resampling : Resampling
        Resampling method, alias of rasterio.warp.Resampling.
    src : RegularGrid, optional
        Definition of the input field grid

    Raises
    ------
    ValueError
        If the input field is not defined on a regular grid in rotated latlon or
        if the input field geography metadata does not have consistent grid parameters.

    Returns
    -------
    xarray.DataArray
        Field regridded in the destination grid.

    """
    if src is None:
        src = RegularGrid.from_field(field)

    def reproject_layer(field):
        output = np.zeros((dst.ny, dst.nx))
        warp.reproject(
            source=field[::-1],
            destination=output,
            src_crs=src.crs,
            src_transform=src.transform,
            dst_crs=dst.crs,
            dst_transform=dst.transform,
            resampling=resampling,
        )
        return output[::-1]

    # output dims renamed to workaround limitation that overlapping dims in the input
    # must not change in size
    data = xr.apply_ufunc(
        reproject_layer,
        field,
        input_core_dims=[["y", "x"]],
        output_core_dims=[["y1", "x1"]],
        vectorize=True,
    ).rename({"x1": "x", "y1": "y"})

    attrs = field.attrs
    if md := _get_metadata(dst):
        attrs = attrs | metadata.override(field.metadata, **md)

    return xr.DataArray(data, attrs=attrs)


def _icon2regular(
    field: xr.DataArray, dst: RegularGrid, indices: np.ndarray, weights: np.ndarray
) -> xr.DataArray:
    mask = np.all(indices != 0, axis=-1)

    def reproject_layer(field):
        out_shape = field.shape[:-1] + (dst.ny, dst.nx)
        values = np.take(field, indices, axis=-1)
        if np.any(np.isnan(values)):
            warnings.warn("Interpolation of missing values is not supported.")
        vmin = np.min(values, axis=-1)
        vmax = np.max(values, axis=-1)
        result = np.einsum("...ij,ij->...i", values, weights)
        masked = np.where(mask, result, np.nan)
        return np.clip(masked, vmin, vmax).reshape(out_shape)

    data = xr.apply_ufunc(
        reproject_layer,
        field,
        input_core_dims=[["cell"]],
        output_core_dims=[["y", "x"]],
    )

    attrs = field.attrs
    if md := _get_metadata(dst):
        attrs = attrs | metadata.override(field.metadata, **md)

    return xr.DataArray(data, attrs=attrs)


def icon2geolatlon(field: xr.DataArray) -> xr.DataArray:
    """Remap ICON native grid data to the geolatlon grid.

    The interpolation is done with pre-computed weights based on the RBF
    interpolation method as implemented in icon-tools from the DWD.

    Parameters
    ----------
    field : xarray.DataArray
        A field with data in the ICON native grid.

    Returns
    -------
    xarray.DataArray
        Field with data remapped to the geolatlon grid.

    """
    gid = field.metadata.get("uuidOfHGrid")
    coeffs = icon_grid.get_remap_coeffs(gid, "geolatlon")
    indices = coeffs["rbf_B_glbidx"].values
    weights = coeffs["rbf_B_wgt"].values

    dst = RegularGrid(
        crs=CRS.from_string("epsg:4326"),
        nx=coeffs.nx,
        ny=coeffs.ny,
        xmin=coeffs.xmin,
        ymin=coeffs.ymin,
        xmax=coeffs.xmax,
        ymax=coeffs.ymax,
    )

    return _icon2regular(field, dst, indices, weights)


def icon2rotlatlon(field: xr.DataArray) -> xr.DataArray:
    """Remap ICON native grid data to the rotated latlon grid.

    The interpolation is done with pre-computed weights based on the RBF
    interpolation method as implemented in icon-tools from the DWD.

    Parameters
    ----------
    field : xarray.DataArray
        A field with data in the ICON native grid.

    Returns
    -------
    xarray.DataArray
        Field with data remapped to the rotated latlon grid.

    """
    gid = field.metadata.get("uuidOfHGrid")
    coeffs = icon_grid.get_remap_coeffs(gid, "rotlatlon")
    indices = coeffs["rbf_B_glbidx"].values
    weights = coeffs["rbf_B_wgt"].values

    geo = {
        "gridType": "rotated_ll",
        "longitudeOfSouthernPoleInDegrees": coeffs.north_pole_lon - 180,
        "latitudeOfSouthernPoleInDegrees": -1 * coeffs.north_pole_lat,
    }
    dst = RegularGrid(
        crs=_get_crs(geo),
        nx=coeffs.nx,
        ny=coeffs.ny,
        xmin=coeffs.xmin,
        ymin=coeffs.ymin,
        xmax=coeffs.xmax,
        ymax=coeffs.ymax,
    )

    return _icon2regular(field, dst, indices, weights)


def _linear_weights(pts_src: ArrayLike, pts_dst: ArrayLike) -> tuple[NDArray, NDArray]:
    """Compute indices and weights for barycentric linear interpolation."""
    tri = Delaunay(pts_src)
    simplex = tri.find_simplex(pts_dst)
    isfound = simplex != -1
    vertices = np.take(tri.simplices, simplex, axis=0)
    indices = np.where(isfound[:, None], vertices, 0)

    # note that zero is a valid index
    # however, a full line of zeros is interpreted as out of bounds
    # in the interpolation step

    temp = np.take(tri.transform, simplex, axis=0)
    delta = pts_dst - temp[:, 2]
    bary = np.einsum("njk,nk->nj", temp[:, :2, :], delta)
    wgts = np.hstack((bary, 1 - bary.sum(axis=1, keepdims=True)))
    weights = np.where(isfound[:, None], wgts, 0)

    return indices, weights


def _linear_weights_cropped_domain(
    pts_src: NDArray, pts_dst: NDArray, buffer: float = 4e3
) -> tuple[NDArray, NDArray]:
    """Compute linear interpolation weights from a cropped source grid.

    Crops the source grid to a box with a buffer outside of the destination grid. The
    default buffer is 4 km when using a meter-based coordinate system.

    Both pts_src and pts_dst are expected to be rank 2 arrays representing a list of
    points using the same coordinate system.
    """
    xmin, ymin = np.min(pts_dst, axis=0) - buffer
    xmax, ymax = np.max(pts_dst, axis=0) + buffer
    x, y = np.transpose(pts_src)
    mask = (xmin < x) & (x < xmax) & (ymin < y) & (y < ymax)
    [idx] = np.nonzero(mask)
    indices, weights = _linear_weights(pts_src[idx], pts_dst)
    return idx[indices], weights


def iconremap(
    field: xr.DataArray, dst: RegularGrid, method: Literal["byc"] = "byc"
) -> xr.DataArray:
    """Remap ICON native grid data to a regular grid.

    Note that the interpolation method is linear.

    Parameters
    ----------
    field : xarray.DataArray
        A field with data in the ICON native grid.
    dst : RegularGrid
        A regular grid in any coordinate system.
    method : Literal["byc"]
        Method used to perform the interpolation.

        Available methods:
        - byc: Barycentric linear interpolation.

    Returns
    -------
    xarray.DataArray
        Field with data remapped to the given swiss grid.

    """
    if method not in {"byc"}:
        raise NotImplementedError(f"method: {method} is not implemented")

    utm_crs = "epsg:32632"  # UTM zone 32N

    transformer_src = Transformer.from_crs("epsg:4326", utm_crs, always_xy=True)
    points_src = transformer_src.transform(field.lon, field.lat)

    gx, gy = np.meshgrid(dst.x, dst.y)
    transformer_dst = Transformer.from_crs(dst.crs.wkt, utm_crs, always_xy=True)
    points_dst = transformer_dst.transform(gx.flat, gy.flat)

    xy = np.array(points_src).T
    uv = np.array(points_dst).T

    indices, weights = _linear_weights_cropped_domain(xy, uv)

    transformer_geo = Transformer.from_crs(dst.crs.wkt, "epsg:4326", always_xy=True)
    lon, lat = transformer_geo.transform(gx, gy)

    return _icon2regular(field, dst, indices, weights).assign_coords(
        lon=(("y", "x"), lon), lat=(("y", "x"), lat)
    )



"""Manage GRIB metadata."""

# Standard library
import dataclasses as dc
import logging
import typing

# Third-party
import numpy as np
import xarray as xr
from earthkit.data.core.metadata import Metadata  # type: ignore

# Local
from . import grib_decoder

_logger = logging.getLogger(__name__)


VCOORD_TYPE = {
    "generalVertical": ("model_level", -0.5),
    "generalVerticalLayer": ("model_level", 0.0),
    "isobaricInPa": ("pressure", 0.0),
}


def extract(metadata: Metadata) -> dict[str, typing.Any]:
    if metadata.get("gridType") == "unstructured_grid":
        vref_flag = False
    else:
        [vref_flag] = grib_decoder.get_code_flag(
            metadata.get("resolutionAndComponentFlags"), [5]
        )

    level_type = metadata.get("typeOfLevel")
    vcoord_type, zshift = VCOORD_TYPE.get(level_type, (level_type, 0.0))

    return {
        "parameter": metadata.as_namespace("parameter"),
        "geography": metadata.as_namespace("geography"),
        "vref": "native" if vref_flag else "geo",
        "vcoord_type": vcoord_type,
        "origin_z": zshift,
    }


def override(metadata: Metadata, **kwargs: typing.Any) -> dict[str, typing.Any]:
    """Override GRIB metadata.

    Note that no special consideration is made for maintaining consistency when
    overriding template definition keys such as productDefinitionTemplateNumber.
    Note that the origin components in x and y are left untouched.

    Parameters
    ----------
    metadata : Metadata
        Metadata of the input GRIB metadata
    kwargs : Any
        Keyword arguments forwarded to earthkit-data GribMetadata override method

    Returns
    -------
    dict[str, Any]
        Updated metadata along with the geography and parameter namespaces

    """
    if metadata["editionNumber"] == 1:
        return {
            "metadata": metadata,
            **extract(metadata),
        }

    md = metadata.override(**kwargs)

    return {
        "metadata": md,
        **extract(md),
    }


@dc.dataclass
class Grid:
    """Coordinates of the reference grid.

    Attributes
    ----------
    lon_first_grid_point: float
        longitude of first grid point in rotated lat-lon CRS
    lat_first_grid_point: float
        latitude of first grid point in rotated lat-lon CRS

    """

    lon_first_grid_point: float
    lat_first_grid_point: float


def load_grid_reference(metadata: Metadata) -> Grid:
    """Construct a grid from a reference parameter.

    Parameters
    ----------
    metadata : Metadata
        GRIB metadata defining the reference grid.

    Returns
    -------
    Grid
        reference grid

    """
    return Grid(
        metadata["longitudeOfFirstGridPointInDegrees"],
        metadata["latitudeOfFirstGridPointInDegrees"],
    )


def compute_origin(ref_grid: Grid, field: xr.DataArray) -> dict[str, float]:
    """Compute horizontal components of the origin dict.

    Parameters
    ----------
    ref_grid : Grid
        reference grid
    field : xarray.DataArray
        field for which to compute the origin

    Returns
    -------
    dict[str, float]
        Horizontal components of the origin

    """
    x0 = ref_grid.lon_first_grid_point % 360
    y0 = ref_grid.lat_first_grid_point
    geo = field.geography
    dx = geo["iDirectionIncrementInDegrees"]
    dy = geo["jDirectionIncrementInDegrees"]
    x0_key = "longitudeOfFirstGridPointInDegrees"
    y0_key = "latitudeOfFirstGridPointInDegrees"

    return {
        "origin_x": np.round((geo[x0_key] % 360 - x0) / dx, 1),
        "origin_y": np.round((geo[y0_key] - y0) / dy, 1),
    }


def _uses_icon_grid(metadata: Metadata) -> bool:
    """Determine if the data is on a MeteoSwiss ICON grid.

    Parameters
    ----------
    metadata : Metadata
        GRIB metadata containing the grid definition.

    Returns
    -------
    bool
        True if the data was created by an MCH ICON forecast.

    """
    return (
        metadata.get("centre", default="") == "lssw"
        and (
            metadata.get("generatingProcessIdentifier", default=0) == 141
            or metadata.get("generatingProcessIdentifier", default=0) == 142
        )
        and metadata.get("gridType") == "unstructured_grid"
    )


def set_origin_xy(ds: dict[str, xr.DataArray], ref_param: str) -> None:
    """Set horizontal components of the origin attribute.

    Parameters
    ----------
    ds : dict[str, xarray.DataArray]
        Dataset of fields to update.
    ref_param : str
        Name of the parameter field to use as a reference. Must be a key of ds.

    Raises
    ------
    KeyError
        if the ref_param key is not found in the input dataset

    """
    if ref_param not in ds:
        raise KeyError(f"ref_param {ref_param} not present in dataset.")

    if _uses_icon_grid(ds[ref_param].metadata):
        _logger.warning(
            "Data is on the ICON grid, not setting origin values. "
            "Setting the origin components is intended for support with horizontal "
            "grid staggering, which is not used in the output of the ICON model."
        )
        return

    ref_grid = load_grid_reference(ds[ref_param].metadata)
    for field in ds.values():
        field.attrs |= compute_origin(ref_grid, field)


def extract_pv(metadata: Metadata) -> dict[str, xr.DataArray]:
    """Extract hybrid level coefficients.

    Parameters
    ----------
    metadata : Metadata
        GRIB metadata containing the pv metadata.

    Returns
    -------
    dict[str, xarray.DataArray]
        Hybrid level coefficients.

    """
    pv = metadata.get("pv")

    if pv is None:
        return {}

    i = len(pv) // 2
    return {
        "ak": xr.DataArray(pv[:i], dims="z"),
        "bk": xr.DataArray(pv[i:], dims="z"),
    }


def extract_hcoords(metadata: Metadata) -> dict[str, xr.DataArray]:
    """Extract horizontal coordinates.

    Parameters
    ----------
    metadata : Metadata
        GRIB metadata containing the grid definition.

    Returns
    -------
    dict[str, xarray.DataArray]
        Horizontal coordinates in geolatlon.

    """
    geo = metadata.geography
    return {
        "lat": xr.DataArray(dims=("y", "x"), data=geo.latitudes().reshape(geo.shape())),
        "lon": xr.DataArray(
            dims=("y", "x"), data=geo.longitudes().reshape(geo.shape())
        ),
    }


def is_staggered_horizontal(field: xr.DataArray) -> bool:
    """Determine if the field is on a staggered horizontal grid.

    Parameters
    ----------
    field: xr.DataArray
        Field containing the grid definition.

    Raises
    ------
    ValueError
        if the field is a on regular grid without origin_x and origin_y set.

    Returns
    -------
    bool
        True if the field is on a staggered horizontal grid.

    """
    if _uses_icon_grid(field.metadata):
        return False

    if "origin_x" not in field.attrs or "origin_y" not in field.attrs:
        raise ValueError("Field is missing origin, run set_origin_xy on the data set.")
    return field.origin_x != 0.0 or field.origin_y != 0.0



"""Helper class to cache data.

The helper is intended to create a cache of grib files with a layout that is
suitable for reading by fieldextra.

"""

# Standard library
import dataclasses as dc
from itertools import product
from pathlib import Path

# Local
from . import data_source

DEFAULT_FILES = {
    "inputi": "<mmm>/lfff<ddhh>0000",
    "inputc": "<mmm>/lfff00000000c",
}


@dc.dataclass
class DataCache:
    cache_dir: Path
    fields: dict[str, list]
    files: dict[str, str] = dc.field(default_factory=lambda: DEFAULT_FILES)
    steps: list[int] = dc.field(default_factory=lambda: [0])
    numbers: list[int] = dc.field(default_factory=lambda: [0])
    _populated: list[Path] = dc.field(default_factory=list, init=False)

    def __post_init__(self):
        if not self.fields.keys() <= self.files.keys():
            raise ValueError("fields keys must be a subset of files keys")

    @property
    def populated_files(self) -> list[Path]:
        return self._populated

    @property
    def conf_files(self) -> dict[str, Path]:
        return {
            label: self.cache_dir / pattern for label, pattern in self.files.items()
        }

    def _iter_files(self):
        # support more patterns ?
        # https://github.com/COSMO-ORG/fieldextra/blob/develop/documentation/README.user#L2797
        patterns = (
            ("<mmm>", "{mmm:03d}"),
            ("<ddhh>", "{dd:02d}{hh:02d}"),
        )
        for label, name in self.files.items():
            name = name.lower()
            numbers = self.numbers if "<mmm>" in name else [None]
            steps = self.steps if "<ddhh>" in name else [None]
            for src, dst in patterns:
                name = name.replace(src, dst)
            for number, step in product(numbers, steps):
                dd = step // 24 if step is not None else None
                hh = step % 24 if step is not None else None
                yield label, name.format(mmm=number, dd=dd, hh=hh), number, step

    def _iter_requests(self, label: str, number: int | None, step: int | None):
        param_map: dict[str, list[str]] = {}
        for param, levtype in self.fields[label]:
            param_map.setdefault(levtype, []).append(param)

        for levtype, params in param_map.items():
            req = {"param": params, "levtype": levtype}
            if number is not None:
                req["number"] = number
            if step is not None:
                req["step"] = step
            yield req

    def populate(self, source: data_source.DataSource):
        for label, rel_path, number, step in self._iter_files():
            path = self.cache_dir / rel_path
            path.parent.mkdir(parents=True, exist_ok=True)

            with path.open("ba") as f:
                for req in self._iter_requests(label, number, step):
                    for field in source.retrieve(req):
                        f.write(field.message())

            self._populated.append(path)

    def clear(self):
        for path in self._populated:
            path.unlink()



"""Horizontal cropping operator."""

# Standard library
import typing

# Third-party
import numpy as np
import xarray as xr

# Local
from .. import metadata
from . import gis


class Bounds(typing.NamedTuple):
    xmin: int
    xmax: int
    ymin: int
    ymax: int


def crop(field: xr.DataArray, bounds: Bounds) -> xr.DataArray:
    """Crop the field to the given bounds.

    Only fields defined on regular grids in rotlatlon coordinates,
    without rotation nor flipped axes are supported.

    Parameters
    ----------
    field : xarray.DataArray
        The field to crop.
    bounds : Bounds
        Bounds of the cropped area given as indices of the array in following order:
        xmin, xmax, ymin, ymax.
        All bounds are inclusive.

    Raises
    ------
    ValueError
        If there are any consistency issues with the provided bounds
        or any of the conditions on the input grid not met.

    Returns
    -------
    xarray.DataArray
        The data is set to cropped domain and the metadata is updated accordingly.

    """
    xmin, xmax, ymin, ymax = bounds

    sizes = field.sizes
    if (
        xmin > xmax
        or ymin > ymax
        or any(v < 0 for v in bounds)
        or xmax >= sizes["x"]
        or ymax >= sizes["y"]
    ):
        raise ValueError(f"Inconsistent bounds: {bounds}")

    grid = gis.get_grid(field.geography)
    lon_min, lon_max = np.round(grid.rlon.isel(x=[xmin, xmax]).values * 1e6)
    lat_min, lat_max = np.round(grid.rlat.isel(y=[ymin, ymax]).values * 1e6)
    ni = xmax - xmin + 1
    nj = ymax - ymin + 1
    npts = ni * nj

    return xr.DataArray(
        field.isel(x=slice(xmin, xmax + 1), y=slice(ymin, ymax + 1)),
        attrs=metadata.override(
            field.metadata,
            longitudeOfFirstGridPoint=lon_min,
            longitudeOfLastGridPoint=lon_max,
            Ni=ni,
            latitudeOfFirstGridPoint=lat_min,
            latitudeOfLastGridPoint=lat_max,
            Nj=nj,
            numberOfDataPoints=npts,
        ),
    )



"""Physical constants."""

g = 9.80665  # [m s-2]
omega = 7.2921e-5
earth_radius = 6371229.0  # value from COSMO GRIB2 key

r_d = 287.05  # Specific gas constant for dry air [J kg-1 K-1]
r_v = 461.51  # Gas constant for water vapour[J kg-1 K-1]
rdv = r_d / r_v
o_rdv = 1.0 - rdv
cp_d = 1005.0  # Specific heat capacity at constant pressure for dry air [J kg-1 K]
rvd = r_v / r_d
rdocp = r_d / cp_d
rvd_o = rvd - 1.0

# saturation vapour pressure (Tetens's formula, see
# http://www.ecmwf.int/sites/default/files/elibrary/2015/9208-part-i-observation-processing.pdf)
b1 = 611.21  # Pressure at triple point of water [Pa]
b2w = 17.502  # over water
b2i = 22.587  # over ice
b3 = 273.16  # Temperature at triple point of water [K]
b4w = 32.19  # [K] over water
b4i = -0.7  # [K] over ice

# Surface pressure reference for omega slope
surface_pressure_ref = 101325.0  # [Pa]

# Radiation
emissivity_surface = 0.996  # [-]
boltzman_cst = 5.6697e-8  # [W m-2 K-4]



"""Product base classes."""

# Standard library
import dataclasses as dc
from abc import ABCMeta, abstractmethod
from collections import Counter
from typing import NamedTuple

# Local
from . import data_source, grib_decoder, tasking


class Request(NamedTuple):
    param: str
    levtype: str = "ml"
    levelist: tuple[int, ...] | None = None


@dc.dataclass
class ProductDescriptor:
    input_fields: dict[str, Request]


class Product(metaclass=ABCMeta):
    """Base class for products."""

    def __init__(
        self,
        input_fields: dict[str, Request],
        delay_entire_product: bool = False,
    ):
        self._desc = ProductDescriptor(input_fields=input_fields)
        self._delay_entire_product = delay_entire_product

    @abstractmethod
    def _run(self, *args, **kwargs):
        pass

    def __call__(self, *args, **kwargs):
        if self._delay_entire_product:
            return tasking.delayed(self._run)(*args, **kwargs)
        else:
            return self._run(*args, **kwargs)

    @property
    def descriptor(self):
        return self._desc


def _merge_requests(
    products: list[Product],
) -> tuple[dict[str, Request], dict[str, list[str]]]:
    """Deduplicate field requests for a collection of products.

    Raises RuntimeError if there are conflicts meaning that there
    different requests assigned to the same name.

    Returns the merged mapping of names to requests and name aliases.
    The alias is needed because a product will expect to find a requested
    field under the name that was attached to the request.
    """
    req_labels: dict[Request, set[str]] = {}
    for product in products:
        for name, req in product.descriptor.input_fields.items():
            name_set = req_labels.setdefault(req, set())
            name_set.add(name)

    # check for any conflicts in the names
    cnt = Counter(name for names in req_labels.values() for name in names)
    conflicts = [name for name, count in cnt.items() if count > 1]
    if conflicts:
        raise RuntimeError(f"These names are involved in conflicts: {conflicts}")

    result: dict[str, Request] = {}
    aliases: dict[str, list[str]] = {}
    for req, names in req_labels.items():
        name, *alias = names
        result[name] = req
        aliases[name] = alias

    return result, aliases


def run_products(
    products: list[Product],
    source: data_source.DataSource,
    ref_param: Request = Request("HHL"),
):
    """Run multiple products.

    Parameters
    ----------
    products : list[Product]
        List of products to run.
    source : DataSource
        Data source from which to request the fields.
    ref_param : Request, optional
        Reference parameter to determine the origin of the grid. Defaults to HHL.

    Raises
    ------
    RuntimeError
        if a single name is mapped to different input field requests.

    Returns
    -------
    tuple[Any, ...]
        Returns the collection of results.

    """
    reqs, aliases = _merge_requests(products)

    reader = grib_decoder.GribReader(source, ref_param=ref_param._asdict())
    ds = reader.load({key: req._asdict() for key, req in reqs.items()})
    for name, alias in aliases.items():
        ds.update({a: ds[name] for a in alias})

    results = [product(ds) for product in products]

    return tasking.compute(*results)



"""Vertical interpolation operators."""

# Standard library
from typing import Literal, Sequence

# Third-party
import numpy as np
import xarray as xr

# First-party
from meteodatalab.operators.support_operators import (
    TargetCoordinates,
    TargetCoordinatesAttrs,
    init_field_with_vcoord,
)


def interpolate_k2p(
    field: xr.DataArray,
    mode: Literal["linear_in_p", "linear_in_lnp", "nearest_sfc"],
    p_field: xr.DataArray,
    p_tc_values: Sequence[float],
    p_tc_units: Literal["Pa", "hPa"],
) -> xr.DataArray:
    """Interpolate a field from model (k) levels to pressure coordinates.

    Example for vertical interpolation to isosurfaces of a target field,
    which is strictly monotonically decreasing with height.



    Parameters
    ----------
    field : xarray.DataArray
        field to interpolate (only typeOfLevel="generalVerticalLayer" is supported)
    mode : str
        interpolation algorithm, one of {"linear_in_p", "linear_in_lnp", "nearest_sfc"}
    p_field : xarray.DataArray
        pressure field on k levels in Pa
        (only typeOfLevel="generalVerticalLayer" is supported)
    p_tc_values : list of float
        pressure target coordinate values
    p_tc_units : str
        pressure target coordinate units

    Returns
    -------
    field_on_tc : xarray.DataArray
        field on target (i.e., pressure) coordinates

    """
    # TODO: check missing value consistency with GRIB2 (currently comparisons are
    #       done with np.nan)
    #       check that p_field is the pressure field, given in Pa (can only be done
    #       if attributes are consequently set)
    #       check that field and p_field are compatible (have the same
    #       dimensions and sizes)
    #       print warn message if result contains missing values

    # Initializations
    # ... supported interpolation modes
    interpolation_modes = ("linear_in_p", "linear_in_lnp", "nearest_sfc")
    if mode not in interpolation_modes:
        raise RuntimeError("interpolate_k2p: unknown mode", mode)
    # ... supported tc units and corresponding conversion factors to Pa
    p_tc_unit_conversions = dict(Pa=1.0, hPa=100.0)
    if p_tc_units not in p_tc_unit_conversions.keys():
        raise RuntimeError(
            "interpolate_k2p: unsupported value of p_tc_units", p_tc_units
        )
    # ... supported range of pressure tc values (in Pa)
    p_tc_min = 1.0
    p_tc_max = 120000.0

    # Define vertical target coordinates (tc)
    tc_factor = p_tc_unit_conversions[p_tc_units]
    tc_values = np.array(sorted(p_tc_values)) * tc_factor
    if np.any((tc_values < p_tc_min) | (tc_values > p_tc_max)):
        raise RuntimeError(
            "interpolate_k2p: target coordinate value out of range "
            f"(must be in interval [{p_tc_min}, {p_tc_max}]Pa)"
        )
    tc = TargetCoordinates(
        type_of_level="isobaricInPa",
        values=tc_values.tolist(),
        attrs=TargetCoordinatesAttrs(
            units="Pa",
            positive="down",
            standard_name="air_pressure",
            long_name="pressure",
        ),
    )

    # Check that typeOfLevel is supported and equal for both field and p_field
    if field.vcoord_type != "model_level":
        raise RuntimeError(
            "interpolate_k2p: field to interpolate must be defined on model levels"
        )
    if p_field.vcoord_type != "model_level":
        raise RuntimeError(
            "interpolate_k2p: pressure field must be defined on model levels"
        )
    # Check that dimensions are the same for field and p_field
    if field.origin_z != p_field.origin_z:
        raise RuntimeError(
            "interpolate_k2p: field and p_field must have equal vertical staggering"
        )

    # Prepare output field field_on_tc on target coordinates
    field_on_tc = init_field_with_vcoord(field.broadcast_like(p_field), tc, np.nan)

    # Interpolate
    # ... prepare interpolation
    pkm1 = p_field.shift(z=1)
    fkm1 = field.shift(z=1)

    # ... loop through tc values
    for tc_idx, p0 in enumerate(tc_values):
        # ... find the 3d field where pressure is > p0 on level k
        # and was <= p0 on level k-1
        p2 = p_field.where((p_field > p0) & (pkm1 <= p0))
        # ... extract the index k of the vertical layer at which p2 adopts its minimum
        #     (corresponds to search from top of atmosphere to bottom)
        # ... note that if the condition above is not fulfilled, minind will
        # be set to k_top
        minind = p2.fillna(p_tc_max).argmin(dim="z")
        # ... extract pressure and field at level k
        p2 = p2[{"z": minind}]
        f2 = field[{"z": minind}]
        # ... extract pressure and field at level k-1
        # ... note that f1 and p1 are both undefined, if minind equals k_top
        f1 = fkm1[{"z": minind}]
        p1 = pkm1[{"z": minind}]

        # ... compute the interpolation weights
        if mode == "linear_in_p":
            # ... note that p1 is undefined, if minind equals k_top, so ratio will
            # be undefined
            ratio = (p0 - p1) / (p2 - p1)

        if mode == "linear_in_lnp":
            # ... note that p1 is undefined, if minind equals k_top, so ratio will
            #  be undefined
            ratio = (np.log(p0) - np.log(p1)) / (np.log(p2) - np.log(p1))

        if mode == "nearest_sfc":
            # ... note that by construction, p2 is always defined;
            #     this operation sets ratio to 0 if p1 (and by construction also f1)
            #     is undefined; therefore, the interpolation formula below works
            #     correctly also in this case
            ratio = xr.where(np.abs(p0 - p1) >= np.abs(p0 - p2), 1.0, 0.0)

        # ... interpolate and update field_on_tc
        field_on_tc[{"z": tc_idx}] = (1.0 - ratio) * f1 + ratio * f2

    return field_on_tc


def interpolate_k2theta(
    field: xr.DataArray,
    mode: Literal["low_fold", "high_fold", "undef_fold"],
    th_field: xr.DataArray,
    th_tc_values: Sequence[float],
    th_tc_units: Literal["K", "cK"],
    h_field: xr.DataArray,
) -> xr.DataArray:
    """Interpolate a field from model levels to potential temperature coordinates.

       Example for vertical interpolation to isosurfaces of a target field
       that is no monotonic function of height.

    Parameters
    ----------
    field : xarray.DataArray
        field to interpolate (only typeOfLevel="generalVerticalLayer" is supported)
    mode : str
        interpolation algorithm, one of {"low_fold", "high_fold", "undef_fold"}
    th_field : xarray.DataArray
        potential temperature theta on k levels in K
        (only typeOfLevel="generalVerticalLayer" is supported)
    th_tc_values : list of float
        target coordinate values
    th_tc_units : str
        target coordinate units
    h_field : xarray.DataArray
        height on k levels (only typeOfLevel="generalVerticalLayer" is supported)

    Returns
    -------
    field_on_tc : xarray.DataArray
        field on target (i.e., theta) coordinates

    """
    # TODO: check missing value consistency with GRIB2
    #       (currently comparisons are done with np.nan)
    #       check that th_field is the theta field, given in K
    #       (can only be done if attributes are consequently set)
    #       check that field, th_field, and h_field are compatible
    #       print warn message if result contains missing values

    # ATTENTION: the attribute "positive" is not set for generalVerticalLayer
    #            we know that for COSMO it would be defined as positive:"down";
    #            for the time being,
    #            we explicitly use the height field on model mid layer
    #            surfaces as auxiliary field

    # Parameters
    # ... supported folding modes
    folding_modes = ("low_fold", "high_fold", "undef_fold")
    if mode not in folding_modes:
        raise RuntimeError("interpolate_k2theta: unsupported mode", mode)

    # ... supported tc units and corresponding conversion factor to K
    # (i.e. to the same unit as theta); according to GRIB2
    #     isentropic surfaces are coded in K; fieldextra codes
    #     them in cK for NetCDF (to be checked)
    th_tc_unit_conversions = dict(K=1.0, cK=0.01)
    if th_tc_units not in th_tc_unit_conversions.keys():
        raise RuntimeError(
            "interpolate_k2theta: unsupported value of th_tc_units", th_tc_units
        )
    # ... supported range of tc values (in K)
    th_tc_min = 1.0
    th_tc_max = 1000.0
    # ... tc values outside range of meaningful values of height,
    # used in tc interval search (in m amsl)
    h_min = -1000.0
    h_max = 100000.0

    # Define vertical target coordinates
    # Sorting cannot be exploited for optimizations, since theta is
    # not monotonous wrt to height tc values are stored in K
    tc_values = np.array(th_tc_values) * th_tc_unit_conversions[th_tc_units]
    if np.any((tc_values < th_tc_min) | (tc_values > th_tc_max)):
        raise RuntimeError(
            "interpolate_k2theta: target coordinate value "
            f"out of range (must be in interval [{th_tc_min}, {th_tc_max}]K)"
        )
    tc = TargetCoordinates(
        type_of_level="theta",
        values=tc_values.tolist(),
        attrs=TargetCoordinatesAttrs(
            units="K",
            positive="up",
            standard_name="air_potential_temperature",
            long_name="potential temperature",
        ),
    )

    # Check that typeOfLevel is supported and equal for field, th_field, and h_field
    if field.vcoord_type != "model_level" or field.origin_z != 0.0:
        raise RuntimeError(
            "interpolate_k2theta: field to interpolate must "
            "be defined on model_level layers"
        )
    if th_field.vcoord_type != "model_level" or th_field.origin_z != 0.0:
        raise RuntimeError(
            "interpolate_k2theta: theta field must be defined on model_level layers"
        )
    if h_field.vcoord_type != "model_level" or h_field.origin_z != 0.0:
        raise RuntimeError(
            "interpolate_k2theta: height field must be defined on model_level layers"
        )

    # Prepare output field field_on_tc on target coordinates
    field_on_tc = init_field_with_vcoord(field.broadcast_like(th_field), tc, np.nan)

    # Interpolate
    # ... prepare interpolation
    thkm1 = th_field.shift(z=1)
    fkm1 = field.shift(z=1)

    # ... loop through tc values
    for tc_idx, th0 in enumerate(tc.values):
        folding_coord_exception = xr.full_like(h_field[{"z": 0}], False)
        # ... find the height field where theta is >= th0 on level k and was <= th0
        #     on level k-1 or where theta is <= th0 on level k
        #     and was >= th0 on level k-1
        h = h_field.where(
            ((th_field >= th0) & (thkm1 <= th0)) | ((th_field <= th0) & (thkm1 >= th0))
        )
        if mode == "undef_fold":
            # ... find condition where more than one interval is found, which
            # contains the target coordinate value
            folding_coord_exception = xr.where(h.notnull(), 1.0, 0.0).sum(dim=["z"])
            folding_coord_exception = folding_coord_exception.where(
                folding_coord_exception > 1.0
            ).notnull()
        if mode in ("low_fold", "undef_fold"):
            # ... extract the index k of the smallest height at which
            # the condition is fulfilled
            tcind = h.fillna(h_max).argmin(dim="z")
        if mode == "high_fold":
            # ... extract the index k of the largest height at which the condition
            # is fulfilled
            tcind = h.fillna(h_min).argmax(dim="z")

        # ... extract theta and field at level k
        th2 = th_field[{"z": tcind}]
        f2 = field[{"z": tcind}]
        # ... extract theta and field at level k-1
        f1 = fkm1[{"z": tcind}]
        th1 = thkm1[{"z": tcind}]

        # ... compute the interpolation weights
        ratio = xr.where(np.abs(th2 - th1) > 0, (th0 - th1) / (th2 - th1), 0.0)

        # ... interpolate and update field_on_tc
        field_on_tc[{"z": tc_idx}] = xr.where(
            folding_coord_exception, np.nan, (1.0 - ratio) * f1 + ratio * f2
        )

    return field_on_tc


def interpolate_k2any(
    field: xr.DataArray,
    mode: Literal["low_fold", "high_fold"],
    tc_field: xr.DataArray,
    tc: TargetCoordinates,
    h_field: xr.DataArray,
) -> xr.DataArray:
    """Interpolate a field from model levels to coordinates w.r.t. an arbitrary field.

    Example for vertical interpolation to isosurfaces of a target field
    that is no monotonic function of height.

    Parameters
    ----------
    field : xarray.DataArray
        field to interpolate (only typeOfLevel="generalVerticalLayer" is supported)
    mode : str
        interpolation algorithm, one of {"low_fold", "high_fold"}
    tc_field : xarray.DataArray
        target field
        (only typeOfLevel="generalVerticalLayer" is supported)
    tc : TargetCoordinates
        target coordinate definition
    h_field : xarray.DataArray
        height on k levels (only typeOfLevel="generalVerticalLayer" is supported)

    Returns
    -------
    field_on_tc : xarray.DataArray
        field on target coordinates

    """
    modes = ("low_fold", "high_fold")
    if mode not in modes:
        raise ValueError(f"Unsupported mode: {mode}")

    for f in (field, tc_field, h_field):
        if f.vcoord_type != "model_level" or f.origin_z != 0.0:
            raise ValueError("Input fields must be defined on full model levels")

    # ... tc values outside range of meaningful values of height,
    # used in tc interval search (in m amsl)
    h_min = -1000.0
    h_max = 100000.0

    # Prepare output field field_on_tc on target coordinates
    field_on_tc = init_field_with_vcoord(field.broadcast_like(tc_field), tc, np.nan)

    # Interpolate
    # ... prepare interpolation
    tckm1 = tc_field.shift(z=1)
    fkm1 = field.shift(z=1)

    for tc_idx, value in enumerate(tc.values):
        # ... find the height field where target is >= value on level k and was <= value
        #     on level k-1 or where target is <= value on level k
        #     and was >= value on level k-1
        h = h_field.where(
            ((tc_field >= value) & (tckm1 <= value))
            | ((tc_field <= value) & (tckm1 >= value))
        )
        if mode == "low_fold":
            # ... extract the index k of the smallest height at which
            # the condition is fulfilled
            tcind = h.fillna(h_max).argmin(dim="z")
        if mode == "high_fold":
            # ... extract the index k of the largest height at which the condition
            # is fulfilled
            tcind = h.fillna(h_min).argmax(dim="z")

        # ... extract target and field at level k
        t2 = tc_field[{"z": tcind}]
        f2 = field[{"z": tcind}]
        # ... extract target and field at level k-1
        f1 = fkm1[{"z": tcind}]
        t1 = tckm1[{"z": tcind}]

        # ... compute the interpolation weights
        ratio = xr.where(np.abs(t2 - t1) > 0, (value - t1) / (t2 - t1), 0.0)

        # ... interpolate and update field_on_tc
        field_on_tc[{"z": tc_idx}] = (1.0 - ratio) * f1 + ratio * f2

    return field_on_tc



"""global configuration for meteodatalab."""

# Standard library
from contextlib import contextmanager
from typing import Any, Literal

config: dict = {}


@contextmanager
def set_values(config: dict = config, **kwargs):
    """Temporarily set configuration values within a context manager.

    Parameters
    ----------
    config : dict, optional
        dictionary object use to hold the configuration.
        Default will use the config object in this module
    **kwargs :
        the configuration key-value pairs to set.

    """
    record: list[tuple[Literal["insert", "replace"], str, Any]] = []

    for key, value in kwargs.items():
        if key in config:
            record.append(("replace", key, config[key]))
        else:
            record.append(("insert", key, None))

        config[key] = value

    try:
        yield config
    finally:
        for op, key, value in reversed(record):
            d = config
            if op == "replace":
                d[key] = value
            else:  # insert
                d.pop(key, None)


def get(
    key: str,
    default: Any = None,
    config: dict = config,
) -> Any:
    """Get values from global config.

    Parameters
    ----------
    key: str
        specifies the name of the key for which the value is requested
    default: Any
        default value to be returned in case the key does not exist in config
    config: dict, optional
        config object holding the mapping. Default value is the global config

    """
    return config.get(key, default)



"""Flexpart operators."""

# Third-party
import numpy as np

# First-party
from meteodatalab.operators.omega_slope import omega_slope
from meteodatalab.operators.time_operators import time_rate


def fflexpart(ds):
    ds_out = {}
    for field in (
        "u",
        "v",
        "t",
        "q",
        "sp",
        "10u",
        "10v",
        "2t",
        "2d",
        "tcc",
        "sd",
    ):
        ds_out[field] = ds[field].isel(lead_time=slice(1, None))

    ds_out["cp"] = time_rate(ds["cp"], np.timedelta64(1, "h"))
    ds_out["lsp"] = time_rate(ds["lsp"], np.timedelta64(1, "h"))

    ds_out["ssr"] = time_rate(ds["ssr"], np.timedelta64(1, "s"))
    ds_out["sshf"] = time_rate(ds["sshf"], np.timedelta64(1, "s"))
    ds_out["ewss"] = time_rate(ds["ewss"], np.timedelta64(1, "s"))
    ds_out["nsss"] = time_rate(ds["nsss"], np.timedelta64(1, "s"))

    ds_out["omega"] = omega_slope(ds["sp"], ds["etadot"], ds["ak"], ds["bk"]).isel(
        z=slice(39, 137), lead_time=slice(1, None)
    )

    return ds_out



"""Open Government Data API helpers."""

# Standard library
import dataclasses as dc
import datetime as dt
import enum
import hashlib
import logging
import os
import typing
from functools import lru_cache
from pathlib import Path
from urllib.parse import urlparse
from uuid import UUID

# Third-party
import earthkit.data as ekd  # type: ignore
import pydantic
import pydantic.dataclasses as pdc
import xarray as xr

# Local
from . import data_source, grib_decoder, icon_grid, util

API_URL = "https://data.geo.admin.ch/api/stac/v1"

logger = logging.getLogger(__name__)
session = util.init_session(logger)


class Collection(str, enum.Enum):
    ICON_CH1 = "ogd-forecasting-icon-ch1"
    ICON_CH2 = "ogd-forecasting-icon-ch2"


def _forecast_prefix(field_name):
    if field_name in {
        "variable",
        "reference_datetime",
        "perturbed",
        "horizon",
    }:
        return f"forecast:{field_name}"
    return field_name


def _parse_datetime(value: str) -> dt.datetime:
    return dt.datetime.strptime(value, "%Y-%m-%dT%H:%M:%S%z")


@pdc.dataclass(
    frozen=True,
    config=pydantic.ConfigDict(
        use_enum_values=True,
        ser_json_timedelta="iso8601",
        alias_generator=pydantic.AliasGenerator(serialization_alias=_forecast_prefix),
    ),
)
class Request:
    collection: Collection = dc.field(metadata=dict(exclude=True))
    variable: str
    reference_datetime: str
    perturbed: bool
    horizon: dt.timedelta

    if typing.TYPE_CHECKING:
        # https://github.com/pydantic/pydantic/issues/10266
        def __init__(self, *args: typing.Any, **kwargs: typing.Any): ...

    @pydantic.computed_field  # type: ignore[misc]
    @property
    def collections(self) -> list[str]:
        return ["ch.meteoschweiz." + str(self.collection)]

    @pydantic.field_validator("reference_datetime", mode="wrap")
    @classmethod
    def valid_reference_datetime(
        cls, input_value: typing.Any, handler: pydantic.ValidatorFunctionWrapHandler
    ) -> str:
        if isinstance(input_value, dt.datetime):
            if input_value.tzinfo is None:
                logger.warn("Converting naive datetime from local time to UTC")
            fmt = "%Y-%m-%dT%H:%M:%SZ"  # Zulu isoformat
            return input_value.astimezone(dt.timezone.utc).strftime(fmt)

        value = handler(input_value)
        parts = value.split("/")
        match parts:
            case [v, ".."] | ["..", v] | [v]:
                # open ended or single value
                _parse_datetime(v)
            case [v1, v2]:
                # range
                d1 = _parse_datetime(v1)
                d2 = _parse_datetime(v2)
                if d2 < d1:
                    raise ValueError("reference_datetime bounds inverted")
            case _:
                raise ValueError(f"Unable to parse reference_datetime: {value}")
        return value

    def dump(self):
        root = pydantic.RootModel(self)
        return root.model_dump(mode="json", by_alias=True)


def _search(url: str, request: Request):
    response = session.post(url, json=request.dump())
    response.raise_for_status()

    obj = response.json()
    result = []
    for item in obj["features"]:
        for asset in item["assets"].values():
            result.append(asset["href"])

    for link in obj["links"]:
        if link["rel"] == "next":
            result.extend(_search(link["href"], request))

    return result


def get_asset_url(request: Request):
    """Get asset URL from OGD.

    The request attributes define filters for the STAC search API according
    to the forecast extension.

    Parameters
    ----------
    request : Request
        Asset search filters, must select a single asset.

    Raises
    ------
    ValueError
        when the request does not select exactly one asset.

    Returns
    -------
    str
        URL of the selected asset.

    """
    result = _search(f"{API_URL}/search", request)
    [asset_url] = result  # expect only one asset

    return asset_url


@lru_cache
def _get_collection_assets(collection_id: str):
    url = f"{API_URL}/collections/{collection_id}/assets"

    response = session.get(url)
    response.raise_for_status()

    return {asset["id"]: asset for asset in response.json().get("assets", [])}


def get_collection_asset_url(collection_id: str, asset_id: str) -> str:
    """Get collection asset URL from OGD.

    Query the STAC collection assets and return the URL for the given asset ID.

    Parameters
    ----------
    collection_id : str
        Full STAC collection ID
    asset_id : str
        The ID of the static asset to retrieve.

    Returns
    -------
    str
        The pre-signed URL of the requested static asset.

    Raises
    ------
    KeyError
        If the asset is not found in the collection.

    """
    assets = _get_collection_assets(collection_id)
    asset_info = assets.get(asset_id)

    if not asset_info or "href" not in asset_info:
        raise KeyError(f"Asset '{asset_id}' not found in collection '{collection_id}'.")

    return asset_info["href"]


def _get_geo_coord_url(uuid: UUID) -> str:
    if (var := os.environ.get("MDL_GEO_COORD_URL")) is not None:
        return var

    model_name = icon_grid.GRID_UUID_TO_MODEL.get(uuid)
    if model_name is None:
        raise KeyError("Grid UUID not found")

    base_model = model_name.removesuffix("-eps")
    collection_id = f"ch.meteoschweiz.ogd-forecasting-{base_model}"
    asset_id = f"horizontal_constants_{model_name}.grib2"

    return get_collection_asset_url(collection_id, asset_id)


def _no_coords(uuid: UUID) -> dict[str, xr.DataArray]:
    return {}


def _geo_coords(uuid: UUID) -> dict[str, xr.DataArray]:
    url = _get_geo_coord_url(uuid)
    source = data_source.URLDataSource(urls=[url])
    ds = grib_decoder.load(source, {"param": ["CLON", "CLAT"]}, geo_coords=_no_coords)
    return {"lon": ds["CLON"].squeeze(), "lat": ds["CLAT"].squeeze()}


def get_from_ogd(request: Request) -> xr.DataArray:
    """Get data from OGD.

    The request attributes define filters for the STAC search API according
    to the forecast extension. It is recommended to enable caching through
    earthkit-data. A warning message is emitted if the cache is disabled.

    Parameters
    ----------
    request : Request
        Asset search filters, must select a single asset.

    Raises
    ------
    ValueError
        when the request does not select exactly one asset.

    Returns
    -------
    xarray.DataArray
        A data array of the selected asset including GRIB metadata and coordinates.

    """
    if ekd.settings.get("cache-policy") == "off":
        doc = "https://earthkit-data.readthedocs.io/en/latest/examples/cache.html"
        logger.warning("Earthkit-data caching is recommended. See: %s", doc)

    asset_url = get_asset_url(request)

    source = data_source.URLDataSource(urls=[asset_url])
    return grib_decoder.load(
        source,
        {"param": request.variable},
        geo_coords=_geo_coords,
    )[request.variable]


def download_from_ogd(request: Request, target: Path) -> None:
    """Download forecast asset and its static coordinate files from OGD.

    The request attributes define filters for the STAC search API according
    to the forecast extension.

    In addition to the main asset, this function downloads static files
    with horizontal and vertical coordinates, as the forecast item
    does not include the horizontal or vertical coordinates.

    Parameters
    ----------
    request : Request
        Asset search filters, must select a single asset.
    target : Path
        Target path where to save the asset, must be a directory.

    Raises
    ------
    ValueError
        when the request does not select exactly one asset.
    RuntimeError
        if the checksum verification fails.

    """
    if target.exists() and not target.is_dir():
        raise ValueError(f"target: {target} must be a directory")

    if not target.exists():
        target.mkdir(parents=True)

    # Download main forecast asset
    asset_url = get_asset_url(request)
    _download_with_checksum(asset_url, target)

    model_suffix = request.collection.removeprefix("ogd-forecasting-")
    collection_id = f"ch.meteoschweiz.{request.collection}"

    # Download coordinate files
    for prefix in ["horizontal", "vertical"]:
        asset_id = f"{prefix}_constants_{model_suffix}-eps.grib2"
        url = get_collection_asset_url(collection_id, asset_id)
        _download_with_checksum(url, target)


def _file_hash(path: Path):
    hash = hashlib.sha256()
    with path.open("rb") as f:
        while chunk := f.read(16 * 1024):
            hash.update(chunk)
    return hash.hexdigest()


def _download_with_checksum(url: str, target: Path) -> None:
    filename = Path(urlparse(url).path).name
    path = target / filename if target.is_dir() else target
    hash_path = path.with_suffix(".sha256")

    if path.exists():
        if hash_path.exists() and hash_path.read_text() == _file_hash(path):
            logger.info(f"File already exists, skipping download: {path}")
            return

    response = session.get(url, stream=True)
    response.raise_for_status()

    hash = response.headers.get("X-Amz-Meta-Sha256")
    if hash is not None:
        hash_path.write_text(hash)

    hasher = hashlib.sha256()
    with path.open("wb") as f:
        for chunk in response.iter_content(16 * 1024):
            f.write(chunk)
            hasher.update(chunk)

    if hash is not None and hash != hasher.hexdigest():
        raise RuntimeError(f"Checksum verification failed for {filename}")



"""Implementation of the potential vorticity operator."""

# Third-party
import numpy as np
import xarray as xr

# Local
from .. import metadata
from .. import physical_constants as pc
from . import diff
from .curl import curl
from .gis import get_grid
from .total_diff import TotalDiff


def compute_pot_vortic(
    u: xr.DataArray,
    v: xr.DataArray,
    w: xr.DataArray,
    theta: xr.DataArray,
    rho_tot: xr.DataArray,
    hhl: xr.DataArray,
) -> xr.DataArray:
    r"""Compute the potential vorticity.

    The potential vorticity is computed with the following formula:

    .. math::
        v_p = \frac{1}{\rho} * \frac{\partial \Theta}{\partial \z} * (c_v + 2 \Omega)

    where
    :math:`\rho` is the total air density,
    :math:`\frac{\partial \Theta}{\partial \z}`
    is the vertical gradient of the potential temperature,
    :math:`c_v` is the curl of the wind in y direction and
    :math`\Omega` is the coriolis term.

    Parameters
    ----------
    u : xarray.DataArray
        Wind in x direction [m/s]
    v : xarray.DataArray
        Wind in y direction [m/s]
    w : xarray.DataArray
        Wind in z direction [m/s]
    theta : xarray.DataArray
        Potential Temperature [K]
    rho_tot : xarray.DataArray
        Total density [kg m-3]
    hhl : xarray.DataArray
        Height at half levels [m]

    Returns
    -------
    xarray.DataArray
        The potential vorticity

    """
    # target coordinates
    deg2rad = np.pi / 180
    lat = (hhl.lat * deg2rad).astype(np.float32)

    rlat = get_grid(hhl.geography).rlat * deg2rad
    total_diff = TotalDiff.from_hhl(hhl)

    # compute curl
    curl1, curl2, curl3 = curl(u, v, w, rlat, total_diff)

    # coriolis terms
    cor2 = 2 * pc.omega / pc.earth_radius * np.cos(lat)
    cor3 = 2 * pc.omega * np.sin(lat)

    dt_dlam = total_diff.d_dlam(diff.dx(theta), diff.dz(theta))
    dt_dphi = total_diff.d_dphi(diff.dy(theta), diff.dz(theta))
    dt_dzeta = total_diff.d_dzeta(diff.dz(theta))

    # potential vorticity
    out = (
        dt_dlam * curl1 + dt_dphi * (curl2 + cor2) - dt_dzeta * (curl3 + cor3)
    ) / rho_tot

    out.attrs = metadata.override(theta.metadata, shortName="POT_VORTIC")

    return out



"""Ninjo product."""

# Standard library
import logging
import sys

# Third-party
import xarray as xr

# First-party
from meteodatalab.operators.destagger import destagger
from meteodatalab.operators.pot_vortic import compute_pot_vortic
from meteodatalab.operators.rho import compute_rho_tot
from meteodatalab.operators.theta import compute_theta
from meteodatalab.operators.vertical_interpolation import (
    interpolate_k2p,
    interpolate_k2theta,
)
from meteodatalab.operators.vertical_reduction import integrate_k

logging.basicConfig(stream=sys.stdout, level=logging.INFO)
logger = logging.getLogger(__name__)


def _compute_pot_vortic(
    U: xr.DataArray,
    V: xr.DataArray,
    W: xr.DataArray,
    T: xr.DataArray,
    P: xr.DataArray,
    QV: xr.DataArray,
    QC: xr.DataArray,
    QI: xr.DataArray,
    HHL: xr.DataArray,
    theta: xr.DataArray,
) -> xr.DataArray:
    logger.info("Computing total density")
    rho_tot = compute_rho_tot(T, P, QV, QC, QI)

    logger.info("Computing potential vorticity")
    return compute_pot_vortic(U, V, W, theta, rho_tot, HHL)


def _compute_mean(
    pot_vortic: xr.DataArray,
    hhl: xr.DataArray,
    hfl: xr.DataArray,
    pressure: xr.DataArray,
) -> xr.DataArray:
    logger.info("Computing mean potential vorticity between 700 and 900 hPa")
    isobars = interpolate_k2p(hfl, "linear_in_lnp", pressure, [700, 900], "hPa")
    h700, h900 = isobars.transpose("z", ...)
    return integrate_k(pot_vortic, "normed_integral", "z2z", hhl, (h900, h700))


def _compute_at_theta(
    theta: xr.DataArray, hfl: xr.DataArray, **fields: xr.DataArray
) -> dict[str, xr.DataArray]:
    logger.info(f"Interpolating {tuple(fields.keys())} at isotherms")
    theta_values = [310.0, 315.0, 320.0, 325.0, 330.0, 335.0]
    return {
        key: interpolate_k2theta(field, "low_fold", theta, theta_values, "K", hfl)
        for key, field in fields.items()
    }


def ninjo_k2th(
    U: xr.DataArray,
    V: xr.DataArray,
    W: xr.DataArray,
    T: xr.DataArray,
    P: xr.DataArray,
    QV: xr.DataArray,
    QC: xr.DataArray,
    QI: xr.DataArray,
    HHL: xr.DataArray,
) -> tuple[xr.DataArray, dict[str, xr.DataArray]]:
    """Compute the Ninjo product.

    Mean potential vorticity is computed between isobaric levels 700hPa and 900hPa.
    The pressure, x and y components of wind velocity and potential vorticity
    are interpolated at potential temperature levels 310K through 335K every 5K.

    Parameters
    ----------
    U: xr.DataArray
        x component of the wind velocity [m/s]
    V: xr.DataArray
        y component of the wind velocity [m/s]
    W: xr.DataArray
        z component of the wind velocity [m/s]
    T: xr.DataArray
        static temperature [K]
    P: xr.DataArray
        static pressure [Pa]
    QV: xr.DataArray
        Specific humidity [kg/kg]
    QC: xr.DataArray
        Specific cloud water content [kg/kg]
    QI: xr.DataArray
        Specific cloud ice content [kg/kg]
    HHL: xr.DataArray
        height of half levels [m AMSL]

    Returns
    -------
    xr.DataArray
        Mean values of potential vorticity between isobaric levels 700hPa and 900hPa
    dict[str, xr.DataArray]
        Values of pressure, horizontal wind velocity and potential vorticity
        interpolated to potential temperature levels 310K through 335K every 5K.

    """
    logger.info("Computing potential temperature")
    theta = compute_theta(P, T)

    pot_vortic = _compute_pot_vortic(U, V, W, T, P, QV, QC, QI, HHL, theta)

    hfl = destagger(HHL, "z")
    output_mean = _compute_mean(pot_vortic, HHL, hfl, P)
    output_at_theta = _compute_at_theta(
        theta,
        hfl,
        p=P,
        u=destagger(U, "x"),
        v=destagger(V, "y"),
        pot_vortic=pot_vortic,
    )

    return output_mean, output_at_theta



"""ICON native grid helper functions."""

# Standard library
from collections.abc import Callable
from functools import partial
from pathlib import Path
from typing import Literal
from uuid import UUID

# Third-party
import numpy as np
import xarray as xr

GRID_UUID_TO_MODEL = {
    UUID("17643da2-5749-59b6-44d2-54a3cd6e2bc0"): "icon-ch1-eps",
    UUID("bbbd5a09-8554-9924-3c7a-4aa4c8762920"): "icon-ch2-eps",
}


def load_grid_from_file(
    uuid: UUID, grid_paths: dict[UUID, Path]
) -> dict[str, xr.DataArray]:
    """Load the clat and clon grid coordinates from .nc format file.

    Parameters
    ----------
    uuid: UUID
        The UUID of the horizontal grid as specified in the GRIB metadata.
    grid_paths: dict[UUID, Path]
        Dictionary mapping from horizontal grid UUID to the path where the data file is.

    Raises
    ------
    KeyError
        If the UUID does not match a known model.

    Returns
    -------
    dict[str, xr.DataArray]
        Dataset containing clon and clat coordinates of the ICON grid cell centers for
        the model.

    """
    grid_path = grid_paths.get(uuid)
    if grid_path is None:
        raise KeyError(
            "No grid file for UUID %s. Known UUIDs are %s.", uuid, grid_paths.keys()
        )
    rad2deg = 180 / np.pi
    ds = xr.open_dataset(grid_path)
    result = ds[["clon", "clat"]].reset_coords() * rad2deg
    return {"lon": result.clon, "lat": result.clat}


def load_grid_from_balfrin() -> Callable[[UUID], dict[str, xr.DataArray]]:
    """Return a grid source to load grid files when running on balfrin."""
    grid_dir = Path("/scratch/mch/jenkins/icon/pool/data/ICON/mch/grids/")
    grid_paths = {
        UUID("17643da2-5749-59b6-44d2-54a3cd6e2bc0"): grid_dir
        / "icon-1/icon_grid_0001_R19B08_mch.nc",
        UUID("bbbd5a09-8554-9924-3c7a-4aa4c8762920"): grid_dir
        / "icon-2/icon_grid_0002_R19B07_mch.nc",
    }
    return partial(load_grid_from_file, grid_paths=grid_paths)


def get_remap_coeffs(
    grid_uuid: str, grid_type: Literal["rotlatlon", "geolatlon"]
) -> xr.Dataset:
    """Get ICON native grid remap indices and weights.

    Parameters
    ----------
    grid_uuid : str
        The UUID of the horizontal grid in hex format.
    grid_type : str
        Type of grid to remap to.

    Raises
    ------
    KeyError
        If the UUID is not found in the GRID_ID constant.

    Returns
    -------
    xarray.Dataset
        Dataset of the remap indices and weights.

    """
    model = GRID_UUID_TO_MODEL[UUID(grid_uuid)]
    coeffs_path = (
        f"/store_new/mch/msopr/icon_workflow_2/iconremap-weights/{model}-{grid_type}.nc"
    )
    return xr.open_dataset(coeffs_path)



"""Util module."""

# Standard library
import logging

# Third-party
import requests
from requests.adapters import HTTPAdapter
from urllib3.util import Retry


def init_session(logger: logging.Logger | None = None) -> requests.Session:
    session = requests.Session()
    retries = Retry(
        total=3,
        allowed_methods=["POST"],
        status_forcelist=[500],
        backoff_factor=0.25,
    )
    session.mount("https://", HTTPAdapter(max_retries=retries))

    if logger is not None:

        def _log_errors(resp: requests.Response, *args, **kwargs):
            if not resp.ok:
                logger.error(
                    "Request to url: %s failed with status code: %d, body: %s",
                    resp.url,
                    resp.status_code,
                    resp.json(),
                )

        session.hooks["response"].append(_log_errors)

    return session



"""Algorithm to compute the density (rho)."""

# Third-party
import xarray as xr

# Local
from .. import metadata
from .. import physical_constants as pc


def compute_rho_tot(
    t: xr.DataArray,
    p: xr.DataArray,
    qv: xr.DataArray,
    qc: xr.DataArray,
    qi: xr.DataArray | None = None,
    qp: xr.DataArray | None = None,
) -> xr.DataArray:
    """Total density of air mixture.

    Assumes perfect gas law, pressure as sum of partial pressures.

    Parameters
    ----------
    t : xarray.DataArray
        Temperature [Kelvin]
    p : xarray.DataArray
        Pressure [Pascal]
    qv : xarray.DataArray
        Specific humidity [kg/kg]
    qc : xarray.DataArray
        Specific cloud water content [kg/kg]
    qi : xarray.DataArray, optional
        Specific cloud ice content [kg/kg]
    qp : xarray.DataArray, optional
        Specific precipitable components content [kg/kg]

    Returns
    -------
    xarray.DataArray
        Total density of air mixture [kg/m**3]

    """
    q = qc
    if qi is not None:
        q += qi
    if qp is not None:
        q += qp

    return xr.DataArray(
        data=p / (pc.r_d * t * (1.0 + pc.rvd_o * qv - q)),
        attrs=metadata.override(p.metadata, shortName="DEN"),
    )



"""Algorithms to support operations on a field."""

# Standard library
import dataclasses as dc
from typing import Any, Literal, Optional, Sequence

# Third-party
import numpy as np
import xarray as xr

# Local
from .. import metadata


@dc.dataclass
class TargetCoordinatesAttrs:
    """Attributes to the target coordinates."""

    standard_name: str
    long_name: str
    units: str
    positive: Literal["up", "down"]


@dc.dataclass
class TargetCoordinates:
    """Target Coordinates."""

    type_of_level: str
    values: Sequence[float]
    attrs: TargetCoordinatesAttrs

    @property
    def size(self):
        return len(self.values)


def init_field_with_vcoord(
    parent: xr.DataArray,
    vcoord: TargetCoordinates,
    fill_value: Any,
    dtype: Optional[np.dtype] = None,
) -> xr.DataArray:
    """Initialize an xarray.DataArray with new vertical coordinates.

    Properties except for those related to the vertical coordinates,
    and optionally dtype, are inherited from the parent xarray.DataArray.

    Parameters
    ----------
    parent : xr.DataArray
        parent field
    vcoord: TargetCoordinates
        target vertical coordinates for the output field
    fill_value : Any
        value the data array of the new field is initialized with
    dtype : np.dtype, optional
        fill value data type; defaults to None (in this case
        the data type is inherited from the parent field)

    Returns
    -------
    xr.DataArray
        new field located at the parent field horizontal coordinates, the target
        coordinates in the vertical and filled with the given value

    """
    # TODO: test that vertical dim of parent is named "generalVerticalLayer"
    # or take vertical dim to replace as argument
    #       be aware that vcoord contains also xr.DataArray GRIB attributes;
    #  one should separate these from coordinate properties
    #       in the interface
    # attrs
    attrs = parent.attrs | metadata.override(
        parent.metadata, typeOfLevel=vcoord.type_of_level
    )
    # dims
    sizes = dict(parent.sizes.items()) | {"z": vcoord.size}
    # coords
    # ... inherit all except for the vertical coordinates
    coords = {c: v for c, v in parent.coords.items() if c != "z"}
    # ... initialize the vertical target coordinates
    coords["z"] = xr.IndexVariable("z", vcoord.values, attrs=dc.asdict(vcoord.attrs))
    # dtype
    if dtype is None:
        dtype = parent.data.dtype

    return xr.DataArray(
        name=parent.name,
        data=np.full(tuple(sizes.values()), fill_value, dtype),
        dims=tuple(sizes.keys()),
        coords=coords,
        attrs=attrs,
    )


def get_grid_coords(n: int, x0: float, dx: float, dim: str) -> xr.DataArray:
    """Compute coordinates for an equally spaced grid.

    Values are rounded to the 6th decimal because the data representation
    in the GRIB specification calls for integers values in microdegrees.
    It is assumed that data input will in the GRIB format and thus subject
    to this property.

    Parameters
    ----------
    n : int
        Number of grid points
    x0 : float
        Coordinates of the origin in the given dimension
    dx : float
        Spacing of the grid along the given dimension
    dim : str
        Dimension along which the grid is defined

    Returns
    -------
    xr.DataArray
        A 1-D field containing the coordinates of the grid along the given dimension

    """
    values = np.arange(n) * dx + x0
    return xr.DataArray(np.round(values, 6), dims=dim)



"""Various time operators."""

# Third-party
import numpy as np
import xarray as xr

# Local
from .. import metadata


def time_rate(var: xr.DataArray, dtime: np.timedelta64):
    """Compute a time rate for a given delta in time.

    It assumes the input data is an accumulated value
    between two time steps of the time coordinate

    Args:
        var: variable that contains the input data
        dtime: delta time of the desired output time rate

    """
    coord = var.valid_time
    result = var.diff(dim="lead_time") / (coord.diff(dim="lead_time") / dtime)
    # No equivalent codes found in Table 10
    # (https://codes.ecmwf.int/grib/format/grib2/ctables/4/10/)
    # For the moment, it is set as 'missing' (255)
    return xr.DataArray(
        data=result,
        attrs=metadata.override(var.metadata, typeOfStatisticalProcessing=255),
    )


def get_nsteps(valid_time: xr.DataArray, dtime: np.timedelta64) -> int:
    """Compute number of steps required for a given time delta.

    Parameters
    ----------
    valid_time : xr.DataArray
        Array of time values.
    dtime : np.timedelta64
        Time difference for which to search the corresponding number of steps.

    Raises
    ------
    ValueError
        if the time difference is not a multiple of the given time index step or
        if the array of time values is not uniform.

    Returns
    -------
    int
        Number of time steps.

    """
    dt = valid_time.diff(dim="lead_time")
    dt0 = dt.values.flat[0]
    uniform = np.all(dt == dt0).item()

    if not uniform:
        msg = "Given field has an irregular time step."
        raise ValueError(msg)

    n = dtime // dt0
    if n * dt0 != dtime:
        msg = "Provided dtime is not a multiple of the time step."
        raise ValueError(msg)

    return n


def delta(field: xr.DataArray, dtime: np.timedelta64) -> xr.DataArray:
    """Compute difference for a given delta in time.

    Parameters
    ----------
    field : xr.DataArray
        Field that contains the input data.
    dtime : np.timedelta64
        Time delta for which to evaluate the difference.

    Raises
    ------
    ValueError
        if dtime is not multiple of the field time step
        or if the time step is not regular.

    Returns
    -------
    xr.DataArray
        The field difference for the given time delta.

    """
    nsteps = get_nsteps(field.valid_time, dtime)
    result = field - field.shift(lead_time=nsteps)
    return xr.DataArray(
        data=result,
        attrs=metadata.override(
            field.metadata,
            lengthOfTimeRange=int(dtime / np.timedelta64(1, "m")),
            indicatorOfUnitForTimeRange=0,
            typeOfStatisticalProcessing=4,
        ),
    )


def resample_average(field: xr.DataArray, dtime: np.timedelta64) -> xr.DataArray:
    """Compute weighted difference for a given delta in time.

    This operator is useful for recomputing time averaged values that are
    aggregated from the reference time to the lead time. The output field
    is averaged with respect to the given time interval for every lead time
    present in the input field. The output for lead times that are smaller
    than the reference time shifted by the interval are undefined.

    Note that this operator is named tdelta in fieldextra.

    Parameters
    ----------
    field : xr.DataArray
        Field that contains the input data.
    dtime : np.timedelta64
        Time delta for which to evaluate the difference.

    Raises
    ------
    ValueError
        if dtime is not multiple of the field time step
        or if the time step is not regular.

    Returns
    -------
    xr.DataArray
        The field weighted difference for the given time delta.

    """
    nsteps = get_nsteps(field.valid_time, dtime)
    weights = (field.valid_time - field.ref_time) / dtime
    weighted = field * weights
    result = weighted - weighted.shift(lead_time=nsteps)
    return xr.DataArray(
        data=result,
        attrs=metadata.override(
            field.metadata,
            lengthOfTimeRange=int(dtime / np.timedelta64(1, "m")),
            indicatorOfUnitForTimeRange=0,
            typeOfStatisticalProcessing=0,
        ),
    )


def resample(field: xr.DataArray, interval: np.timedelta64) -> xr.DataArray:
    """Resample field.

    The interval must be a multiple of the current time step.
    No interpolation is performed.

    Parameters
    ----------
    field : xr.DataArray
        Field that contains the input data.
    interval : np.timedelta64
        Output sample interval.

    Raises
    ------
    ValueError
        if dtime is not multiple of the field time step
        or if the time step is not regular.

    Returns
    -------
    xr.DataArray
        The resampled field.

    """
    nsteps = get_nsteps(field.valid_time, interval)
    return xr.DataArray(
        data=field.sel(lead_time=slice(None, None, nsteps)),
        attrs=metadata.override(
            field.metadata,
            lengthOfTimeRange=int(interval / np.timedelta64(1, "m")),
            indicatorOfUnitForTimeRange=0,
        ),
    )



"""Meteoswiss model data.

Module for retrieving Meteoswiss model data from FDB or Polytope,
and archiving data to FDB.

"""

# Standard library
import io
import logging
import os

# Third-party
import xarray as xr

# Local
from . import data_source, grib_decoder, mars

logger = logging.getLogger(__name__)


def get_from_fdb(request: mars.Request) -> dict[str, xr.DataArray]:
    """Get model data from FDB.

    Parameters
    ----------
    request : mars.Request
        Request for data defined in the mars language.

    Raises
    ------
    RuntimeError
        if the required environment variables for FDB are not set.
    ValueError
        if the request has a feature attribute.

    Returns
    -------
    dict[str, xarray.DataArray]
        Dataset containing the requested data.

    """
    keys = "FDB5_CONFIG", "FDB5_CONFIG_FILE"
    if all(key not in os.environ for key in keys):
        msg = (
            "Required environment variables for FDB are not set."
            "Define one of 'FDB5_CONFIG' or 'FDB5_CONFIG_FILE'"
        )
        logger.exception(msg)
        raise RuntimeError(msg)
    if request.feature is not None:
        msg = "FDB does not support the feature attribute."
        logger.exception(msg)
        raise ValueError(msg)
    logger.info("Getting request %s from FDB.", request)
    source = data_source.FDBDataSource()
    return grib_decoder.load(source, request)


def _polytope_auth_in_env() -> bool:
    polytope_username_auth = "POLYTOPE_USERNAME", "POLYTOPE_PASSWORD"
    return (
        all(env_var not in os.environ for env_var in polytope_username_auth)
        or "POLYTOPE_USER_KEY" in os.environ
    )


def get_from_polytope(request: mars.Request) -> dict[str, xr.DataArray]:
    """Get model data from Polytope.

    Parameters
    ----------
    request : mars.Request
        Request for data defined in the mars language.

    Raises
    ------
    RuntimeError
        if the required environment variables for polytope are not set.

    Returns
    -------
    dict[str, xarray.DataArray]
        Dataset containing the requested data.

    """
    if "POLYTOPE_ADDRESS" not in os.environ or not _polytope_auth_in_env():
        msg = (
            "Required environment variables for polytope are not set."
            "Define 'POLYTOPE_ADDRESS' and set authentication parameters "
            "of either 'POLYTOPE_USERNAME' and 'POLYTOPE_PASSWORD' "
            "or 'POLYTOPE_USER_KEY'."
        )
        logger.exception(msg)
        raise RuntimeError(msg)
    if request.feature is not None:
        source = data_source.PolytopeDataSource(polytope_collection="mchgj")
        [result] = source.retrieve(request)
        return result.to_xarray()
    else:
        collection = "mch"
    logger.info("Getting request %s from polytope collection %s", request, collection)
    source = data_source.PolytopeDataSource(polytope_collection=collection)
    return grib_decoder.load(source, request)


def archive_to_fdb(
    field: xr.DataArray,
    request: mars.Request | None = None,
    bits_per_value: int = 16,
) -> None:
    """Archive a field to FDB.

    Note that all messages will be held in memory during archival.

    Parameters
    ----------
    field : xarray.DataArray
        The field that should be archived.
    request : mars.Request, optional
        The request under which the data should be archived.
        If not provided, the keys are derived from the field.
    bits_per_value : int, optional
        Bits per value encoded in the archived data. (Default: 16)

    """
    try:
        # Third-party
        import pyfdb  # type: ignore
    except ImportError:
        raise ImportError("archive_to_fdb requires the pyfdb package")

    buffer = io.BytesIO()
    grib_decoder.save(field, buffer, bits_per_value)
    fdb = pyfdb.FDB()
    req = request.to_fdb() if request is not None else None
    if request is not None:
        logger.info("Archiving request %s to FDB", request)
    elif "parameter" in field.attrs:
        logger.info("Archiving field %s to FDB", field.parameter["shortName"])
    fdb.archive(buffer.getvalue(), req)



"""Top-level package for meteodata-lab."""

# Standard library
import importlib.metadata

__author__ = "Carlos Osuna"
__email__ = "carlos.osuna@meteoswiss.ch"
__version__ = importlib.metadata.version("meteodata-lab")



"""Wind operators."""

# Third-party
import numpy as np
import xarray as xr

# Local
from ..metadata import is_staggered_horizontal, override
from .gis import vref_rot2geolatlon


def speed(u: xr.DataArray, v: xr.DataArray) -> xr.DataArray:
    """Compute the horizontal wind speed.

    Note that terrain following grid deformation is not accounted for.
    The input fields are required to be located on the mass points.
    The preferred unit is meters per second but any consistent unit system
    acceptable.

    Parameters
    ----------
    u : xarray.DataArray
        the x component of the wind velocity [m/s].
    v : xarray.DataArray
        the y component of the wind velocity [m/s].

    Raises
    ------
    ValueError
        if any of the input fields is located on staggered points.

    Returns
    -------
    xarray.DataArray
        the horizontal wind speed [m/s].

    """
    if is_staggered_horizontal(u) or is_staggered_horizontal(v):
        raise ValueError("The wind components should not be staggered.")

    name = {"U": "SP", "U_10M": "SP_10M"}[u.parameter["shortName"]]
    return xr.DataArray(
        np.sqrt(u**2 + v**2),
        attrs=override(u.metadata, shortName=name),
    )


def direction(u: xr.DataArray, v: xr.DataArray) -> xr.DataArray:
    """Compute the horizontal wind direction.

    Note that terrain following grid deformation is not accounted for.
    The input fields are required to be located on the mass points
    of a grid defined in the rotated or geo latlon coordinate system.
    The preferred unit is meters per second but any consistent unit system
    acceptable.

    Parameters
    ----------
    u : xarray.DataArray
        the x component of the wind velocity [m/s].
    v : xarray.DataArray
        the y component of the wind velocity [m/s].

    Raises
    ------
    ValueError
        if any of the input fields is located on staggered points or on any
        other than a grid in the rotated or geo latlon coordinate system.

    Returns
    -------
    xarray.DataArray
        the horizontal wind direction with respect to geographic North [deg].

    """
    if is_staggered_horizontal(u) or is_staggered_horizontal(v):
        raise ValueError("The wind components should not be staggered.")

    rad2deg = 180 / np.pi
    if u.vref == "native" and v.vref == "native":
        u_g, v_g = vref_rot2geolatlon(u, v)
    elif u.vref == "geo" and v.vref == "geo":
        u_g = u
        v_g = v
    else:
        ValueError(f"Differing or unknown vector references ({u.vref=}, {v.vref=})")
    name = {"U": "DD", "U_10M": "DD_10M"}[u.parameter["shortName"]]
    return xr.DataArray(
        rad2deg * np.arctan2(u_g, v_g) + 180,
        attrs=override(u.metadata, shortName=name),
    )



"""algorithm for destaggering a field."""

# Standard library
from typing import Any, Literal

# Third-party
import numpy as np
import xarray as xr

# Local
from .. import metadata

ExtendArg = Literal["left", "right", "both"] | None


def _intrp_mid(a: np.ndarray) -> np.ndarray:
    return 0.5 * (a[..., :-1] + a[..., 1:])


def _left(a: np.ndarray) -> np.ndarray:
    t = a.copy()
    t[..., 1:] = _intrp_mid(a)
    return t


def _right(a: np.ndarray) -> np.ndarray:
    t = a.copy()
    t[..., :-1] = _intrp_mid(a)
    return t


def _both(a: np.ndarray) -> np.ndarray:
    *m, n = a.shape
    t = np.empty((*m, n + 1))
    t[..., 0] = a[..., 0]
    t[..., -1] = a[..., -1]
    t[..., 1:-1] = _intrp_mid(a)
    return t


def interpolate_midpoint(array: np.ndarray, extend: ExtendArg = None) -> np.ndarray:
    """Interpolate field values onto the midpoints.

    The interpolation is only done on the last dimension of the given array.
    The first or last values can optionally duplicated as per the extend argument.

    Parameters
    ----------
    array : np.ndarray
        Array of field values
    extend : None | Literal["left", "right", "both"]
        Optionally duplicate values on the left, right or both sides.
        Defaults to None.

    Raises
    ------
    ValueError
        If the extend argument is not recognised.

    Returns
    -------
    np.ndarray
        Values of the field interpolated to the midpoint on the last dimension.

    """
    f_map = {
        None: _intrp_mid,
        "left": _left,
        "right": _right,
        "both": _both,
    }
    if extend not in f_map:
        raise ValueError(f"extend arg not in {tuple(f_map.keys())}")
    return f_map[extend](array)


def _update_grid(field: xr.DataArray, dim: Literal["x", "y"]) -> dict[str, Any]:
    geo = field.geography
    if dim == "x":
        lon_min = np.round(geo["longitudeOfFirstGridPointInDegrees"] * 1e6)
        lon_max = np.round(geo["longitudeOfLastGridPointInDegrees"] * 1e6)
        dx = np.round(geo["iDirectionIncrementInDegrees"] * 1e6)
        return metadata.override(
            field.metadata,
            longitudeOfFirstGridPoint=lon_min - dx / 2,
            longitudeOfLastGridPoint=lon_max - dx / 2,
        )
    if dim == "y":
        lat_min = np.round(geo["latitudeOfFirstGridPointInDegrees"] * 1e6)
        lat_max = np.round(geo["latitudeOfLastGridPointInDegrees"] * 1e6)
        dy = np.round(geo["jDirectionIncrementInDegrees"] * 1e6)
        return metadata.override(
            field.metadata,
            latitudeOfFirstGridPoint=lat_min - dy / 2,
            latitudeOfLastGridPoint=lat_max - dy / 2,
        )


def _update_vertical(field) -> dict[str, Any]:
    if field.vcoord_type != "model_level":
        raise ValueError("typeOfLevel must equal generalVertical")
    return metadata.override(
        field.metadata,
        typeOfLevel="generalVerticalLayer",
    )


def destagger(
    field: xr.DataArray,
    dim: Literal["x", "y", "z"],
) -> xr.DataArray:
    """Destagger a field.

    Note that, in the x and y directions, it is assumed that one element
    of the destaggered field is missing on the left side of the domain.
    The first element is thus duplicated to fill the blank.

    Parameters
    ----------
    field : xr.DataArray
        Field to destagger
    dim : Literal["x", "y", "z"]
        Dimension along which to destagger

    Raises
    ------
    ValueError
        Raises ValueError if dim argument is not one of
        {"x","y","z"}.

    Returns
    -------
    xr.DataArray
        destaggered field with dimensions in
        {"x","y","z"}

    """
    dims = list(field.sizes.keys())
    if dim == "x" or dim == "y":
        if field.attrs[f"origin_{dim}"] != 0.5:
            raise ValueError
        attrs = _update_grid(field, dim)
        return (
            xr.apply_ufunc(
                interpolate_midpoint,
                field,
                input_core_dims=[[dim]],
                output_core_dims=[[dim]],
                kwargs={"extend": "left"},
                keep_attrs=True,
            )
            .transpose(*dims)
            .assign_attrs({f"origin_{dim}": 0.0}, **attrs)
            .assign_coords(metadata.extract_hcoords(attrs["metadata"]))
        )
    elif dim == "z":
        if field.origin_z != -0.5:
            raise ValueError
        return (
            xr.apply_ufunc(
                interpolate_midpoint,
                field,
                input_core_dims=[[dim]],
                output_core_dims=[[dim]],
                exclude_dims={dim},
                keep_attrs=True,
            )
            .transpose(*dims)
            .assign_attrs({f"origin_{dim}": 0.0}, **_update_vertical(field))
            .assign_coords(z=field.z[:-1])
        )

    raise ValueError("Unknown dimension", dim)



"""Mars request helper class."""

from __future__ import annotations

# Standard library
import dataclasses as dc
import json
import typing
from collections.abc import Iterable
from enum import Enum
from functools import cache
from importlib.resources import files

# Third-party
import pydantic
import yaml
from pydantic import dataclasses as pdc

ValidationError = pydantic.ValidationError


class Class(str, Enum):
    OPERATIONAL_DATA = "od"


class LevType(str, Enum):
    MODEL_LEVEL = "ml"
    PRESSURE_LEVEL = "pl"
    SURFACE = "sfc"
    SURFACE_OTHER = "sol"
    POT_VORTICITY = "pv"
    POT_TEMPERATURE = "pt"
    DEPTH = "dp"


class Model(str, Enum):
    COSMO_1E = "COSMO-1E"
    COSMO_2E = "COSMO-2E"
    KENDA_1 = "KENDA-1"
    SNOWPOLINO = "SNOWPOLINO"
    ICON_CH1_EPS = "ICON-CH1-EPS"
    ICON_CH2_EPS = "ICON-CH2-EPS"
    KENDA_CH1 = "KENDA-CH1"


class Stream(str, Enum):
    ENS_DATA_ASSIMIL = "enda"
    ENS_FORECAST = "enfo"


class Type(str, Enum):
    DETERMINISTIC = "det"
    ENS_MEMBER = "ememb"
    ENS_MEAN = "emean"
    ENS_STD_DEV = "estdv"


class FeatureType(str, Enum):
    BOUNDINGBOX = "boundingbox"
    TIMESERIES = "timeseries"


class Point(typing.NamedTuple):
    lat: float
    lon: float


@dc.dataclass(frozen=True)
class Range:
    start: int
    end: int
    step: int | None = None


@dc.dataclass(frozen=True)
class BoundingBoxFeature:
    points: list[tuple[float, ...]]
    axes: tuple[str, ...] = ("latitude", "longitude")
    type: typing.Literal[FeatureType.BOUNDINGBOX] = FeatureType.BOUNDINGBOX

    @pydantic.model_validator(mode="after")
    def validate(self):
        min_axes = {"latitude", "longitude"}
        max_axes = {"latitude", "longitude", "level"}
        if len(self.points) != 2:
            raise ValueError("points must contain two points")
        if not (min_axes <= set(self.axes) <= max_axes):
            msg = "axes must contain latitude, longitude and optionally level"
            raise ValueError(msg)
        if any(len(point) != len(self.axes) for point in self.points):
            raise ValueError("points must have same number of components as axes")
        return self


@dc.dataclass(frozen=True)
class TimeseriesFeature:
    points: list[Point]
    time_axis: typing.Literal["step", "date"]
    range: Range | None = None
    axes: tuple[str, str] = ("latitude", "longitude")
    type: typing.Literal[FeatureType.TIMESERIES] = FeatureType.TIMESERIES

    def validate_request(self, request: Request) -> None:
        time_attr = getattr(request, self.time_axis)
        if self.range is not None and time_attr is not None:
            raise ValueError(
                f"only one of {self.time_axis} or feature range must be defined"
            )
        elif self.range is None and time_attr is None:
            raise ValueError(
                f"one of {self.time_axis} or feature range must be defined"
            )


Feature = BoundingBoxFeature | TimeseriesFeature


@cache
def _load_mapping():
    mapping_path = files("meteodatalab.data").joinpath("field_mappings.yml")
    return yaml.safe_load(mapping_path.open())


N_LVL = {
    Model.COSMO_1E: 80,
    Model.COSMO_2E: 60,
    Model.ICON_CH1_EPS: 80,
    Model.ICON_CH2_EPS: 80,
    Model.KENDA_CH1: 80,
}


@pdc.dataclass(
    frozen=True,
    config=pydantic.ConfigDict(use_enum_values=True),
)
class Request:
    param: str | tuple[str, ...]
    date: str | None = None  # YYYYMMDD
    time: str | None = None  # hhmm

    expver: str = "0001"
    levelist: int | tuple[int, ...] | None = None
    number: int | tuple[int, ...] | None = None
    step: int | tuple[int, ...] | None = None

    class_: Class = dc.field(
        default=Class.OPERATIONAL_DATA,
        metadata=dict(alias="class"),
    )
    levtype: LevType = LevType.MODEL_LEVEL
    model: Model = Model.COSMO_1E
    stream: Stream = Stream.ENS_FORECAST
    type: Type = Type.ENS_MEMBER

    feature: Feature | None = dc.field(
        default=None,
        metadata=dict(discriminator="type"),
    )

    @pydantic.model_validator(mode="after")
    def validate(self):
        if isinstance(self.feature, TimeseriesFeature):
            self.feature.validate_request(self)
        return self

    def dump(self):
        if pydantic.__version__.startswith("1"):
            json_str = json.dumps(self, default=pydantic.json.pydantic_encoder)
            obj = json.loads(json_str.replace("class_", "class"))
            return {key: value for key, value in obj.items() if value is not None}

        root = pydantic.RootModel(self)
        return root.model_dump(
            mode="json",
            by_alias=True,
            exclude_none=True,
        )

    def _param_id(self):
        mapping = _load_mapping()
        if isinstance(self.param, Iterable) and not isinstance(self.param, str):
            return [mapping[param]["cosmo"]["paramId"] for param in self.param]
        return mapping[self.param]["cosmo"]["paramId"]

    def _staggered(self):
        mapping = _load_mapping()
        if isinstance(self.param, Iterable) and not isinstance(self.param, str):
            return any(
                mapping[param]["cosmo"].get("vertStag", False) for param in self.param
            )
        return mapping[self.param]["cosmo"].get("vertStag", False)

    def to_fdb(self) -> dict[str, typing.Any]:
        if self.date is None or self.time is None:
            raise RuntimeError("date and time are required fields for FDB.")

        if self.levelist is None and self.levtype == LevType.MODEL_LEVEL:
            n_lvl = N_LVL[self.model]
            if self._staggered():
                n_lvl += 1
            levelist: int | tuple[int, ...] | None = tuple(range(1, n_lvl + 1))
        else:
            levelist = self.levelist

        obj = dc.replace(self, levelist=levelist)
        out = typing.cast(dict[str, typing.Any], obj.dump())
        return out | {"param": self._param_id()}

    def to_polytope(self) -> dict[str, typing.Any]:
        result = self.to_fdb()
        if isinstance(result["param"], list):
            param: str | list[str] = [str(p) for p in result["param"]]
        else:
            param = str(result["param"])
        return result | {"param": param, "model": result["model"].lower()}



"""Decoder for grib data."""

# Standard library
import dataclasses as dc
import datetime as dt
import io
import logging
import typing
from collections import UserDict
from collections.abc import Callable, Mapping, Sequence
from enum import Enum
from itertools import product
from pathlib import Path
from uuid import UUID
from warnings import warn

# Third-party
import earthkit.data as ekd  # type: ignore
import numpy as np
import pandas as pd
import xarray as xr
from numpy.typing import DTypeLike

# Local
from . import data_source, icon_grid, mars, metadata, tasking

logger = logging.getLogger(__name__)

DIM_MAP = {
    "eps": "perturbationNumber",
    "ref_time": "ref_time",
    "lead_time": "step",
    "z": "level",
}
NAME_KEY = "shortName"

GeoCoordsCbk = Callable[[UUID], dict[str, xr.DataArray]]
Request = str | tuple | dict | mars.Request


class ChainMap(UserDict):
    def __init__(self, *maps):
        self._maps = maps

    def __getitem__(self, key):
        for mapping in self._maps:
            try:
                return mapping[key]
            except KeyError:
                pass
        raise KeyError(f"{key} not found")


class GribField(typing.Protocol):
    def metadata(self, *args, **kwargs) -> typing.Any: ...
    def message(self) -> bytes: ...
    def to_numpy(self, dtype: DTypeLike) -> np.ndarray: ...
    def to_latlon(self) -> dict[str, np.ndarray]: ...


class MissingData(RuntimeError):
    pass


class UnitOfTime(Enum):
    MINUTE = 0
    HOUR = 1
    DAY = 2
    SECOND = 13
    MISSING = 255

    @property
    def unit(self):
        if self.name == "MISSING":
            return None
        return self.name.lower()


def _is_ensemble(field) -> bool:
    try:
        return field.metadata("typeOfEnsembleForecast") == 192
    except KeyError:
        return False


def _get_hcoords(
    field: GribField, geo_coords: GeoCoordsCbk | None
) -> tuple[dict[str, xr.DataArray], tuple[str, ...]]:
    hdims: tuple[str, ...]
    if field.metadata("gridType") == "unstructured_grid":
        hdims = ("cell",)
        grid_uuid = UUID(field.metadata("uuidOfHGrid"))
        if geo_coords is None:
            logger.info(
                "No grid source provided when loading data with unstructured grid, "
                "falling back to balfrin grid file locations."
            )
            hcoords = icon_grid.load_grid_from_balfrin()(grid_uuid)
            return hcoords, hdims
        else:
            return geo_coords(grid_uuid), hdims

    hcoords = {
        dim: xr.DataArray(dims=("y", "x"), data=values)
        for dim, values in field.to_latlon().items()
    }
    hdims = ("y", "x")
    return hcoords, hdims


def _parse_datetime(date, time) -> dt.datetime:
    return dt.datetime.strptime(f"{date}{time:04d}", "%Y%m%d%H%M")


def _to_timedelta(value, unit) -> np.timedelta64:
    return pd.to_timedelta(value, unit).to_numpy()


def _get_key(field, dims):
    md = field.metadata()
    step = md["step"]
    unit = "h" if isinstance(step, int) else None
    extra = {
        "ref_time": _parse_datetime(md["dataDate"], md["dataTime"]),
        "step": _to_timedelta(step, unit),
    }
    dim_keys = (DIM_MAP[dim] for dim in dims)
    mapping = ChainMap(extra, md)
    return tuple(mapping[key] for key in dim_keys)


@dc.dataclass
class _FieldBuffer:
    is_ensemble: dc.InitVar[bool]
    dims: tuple[str, ...] = tuple(DIM_MAP)
    hdims: tuple[str, ...] = ("y", "x")
    hcoords: dict[str, xr.DataArray] = dc.field(default_factory=dict)
    metadata: dict[str, typing.Any] = dc.field(default_factory=dict)
    values: dict[tuple[int, ...], np.ndarray] = dc.field(default_factory=dict)

    def __post_init__(self, is_ensemble: bool):
        if not is_ensemble:
            self.dims = self.dims[1:]

    def load(self, field: GribField, geo_coords: GeoCoordsCbk | None) -> None:
        key = _get_key(field, self.dims)
        name = field.metadata(NAME_KEY)
        logger.debug("Received field for param: %s, key: %s", name, key)

        if key in self.values:
            logger.warn("Key collision for param: %s, key: %s", name, key)

        self.values[key] = field.to_numpy(dtype=np.float32)

        if not self.metadata:
            md = field.metadata().override()
            self.metadata = {
                "metadata": md,
                **metadata.extract(md),
            }

        if not self.hcoords:
            self.hcoords, self.hdims = _get_hcoords(field, geo_coords=geo_coords)

    def _gather_coords(self):
        coord_values = zip(*self.values)
        unique = (sorted(set(values)) for values in coord_values)
        coords = {dim: c for dim, c in zip(self.dims, unique)}

        if missing := [
            combination
            for combination in product(*coords.values())
            if combination not in self.values
        ]:
            msg = f"Missing combinations: {missing}"
            logger.exception(msg)
            raise RuntimeError(msg)

        field_shape = next(iter(self.values.values())).shape
        shape = tuple(len(v) for v in coords.values()) + field_shape
        return coords, shape

    def to_xarray(self) -> xr.DataArray:
        if not self.values:
            raise MissingData("No values.")

        coords, shape = self._gather_coords()
        ref_time = xr.DataArray(coords["ref_time"], dims="ref_time")
        lead_time = xr.DataArray(coords["lead_time"], dims="lead_time")
        tcoords = {"valid_time": ref_time + lead_time}

        array = xr.DataArray(
            data=np.array(
                [self.values.pop(key) for key in sorted(self.values)]
            ).reshape(shape),
            coords=coords | self.hcoords | tcoords,
            dims=self.dims + self.hdims,
            attrs=self.metadata,
        )

        if array.vcoord_type != "surface":
            return array

        return array.squeeze("z", drop=True)


def _load_buffer_map(
    source: data_source.DataSource,
    request: Request,
    geo_coords: GeoCoordsCbk | None,
) -> dict[str, _FieldBuffer]:
    logger.info("Retrieving request: %s", request)
    fs = source.retrieve(request)

    buffer_map: dict[str, _FieldBuffer] = {}

    for field in fs:
        name = field.metadata(NAME_KEY)
        if name in buffer_map:
            buffer = buffer_map[name]
        else:
            buffer = buffer_map[name] = _FieldBuffer(_is_ensemble(field))
        buffer.load(field, geo_coords=geo_coords)

    return buffer_map


def load_single_param(
    source: data_source.DataSource,
    request: Request,
    geo_coords: GeoCoordsCbk | None = None,
) -> xr.DataArray:
    """Request data from a data source for a single parameter.

    Parameters
    ----------
    source : data_source.DataSource
        Source to request the data from.
    request : str | tuple[str, str] | dict[str, Any] | meteodatalab.mars.Request
        Request for data from the source in the mars language.
    geo_coords: Callable[[UUID], dict[str, xr.DataArray]] | None
        Callable that returns the horizontal coordinates
        of the grid defined by the given UUID. The dimension must be "cell".

    Raises
    ------
    ValueError
        if more than one param is present in the request.
    RuntimeError
        when all of the requested data is not returned from the data source.

    Returns
    -------
    xarray.DataArray
        A data array of the requested field.

    """
    if (
        isinstance(request, dict)
        and isinstance(request["param"], Sequence)
        and len(request["param"]) > 1
    ):
        raise ValueError("Only one param is supported.")

    buffer_map = _load_buffer_map(source, request, geo_coords)
    [buffer] = buffer_map.values()
    return buffer.to_xarray()


def load(
    source: data_source.DataSource,
    request: Request,
    geo_coords: GeoCoordsCbk | None = None,
) -> dict[str, xr.DataArray]:
    """Request data from a data source.

    Parameters
    ----------
    source : data_source.DataSource
        Source to request the data from.
    request : str | tuple[str, str] | dict[str, Any] | meteodatalab.mars.Request
        Request for data from the source in the mars language.
    geo_coords: Callable[[UUID], dict[str, xr.DataArray]] | None
        Callable that returns the horizontal coordinates
        of the grid defined by the given UUID. The dimension must be "cell".

    Raises
    ------
    RuntimeError
        when all of the requested data is not returned from the data source.

    Returns
    -------
    dict[str, xarray.DataArray]
        A mapping of shortName to data arrays of the requested fields.

    """
    buffer_map = _load_buffer_map(source, request, geo_coords=geo_coords)
    result = {}
    for name, buffer in buffer_map.items():
        try:
            result[name] = buffer.to_xarray()
        except MissingData as e:
            raise RuntimeError(f"Missing data for param: {name}") from e
    return result


class GribReader:
    def __init__(
        self,
        source: data_source.DataSource,
        geo_coords: GeoCoordsCbk | None = None,
        ref_param: Request | None = None,
    ):
        """Initialize a grib reader from a data source.

        Parameters
        ----------
        source : data_source.DataSource
            Data source from which to retrieve the grib fields
        geo_coords: Callable[[UUID], dict[str, xr.DataArray]] | None
            Callable that returns the horizontal coordinates
            of the grid defined by the given UUID. The dimension must be "cell".
        ref_param : str
            name of parameter used to construct a reference grid

        Raises
        ------
        ValueError
            if the grid can not be constructed from the ref_param

        """
        self.data_source = source
        self.geo_coords = geo_coords
        if ref_param is not None:
            warn("GribReader: ref_param is deprecated.")

    @classmethod
    def from_files(
        cls,
        datafiles: list[Path],
        geo_coords: GeoCoordsCbk | None = None,
        ref_param: Request | None = None,
    ):
        """Initialize a grib reader from a list of grib files.

        Parameters
        ----------
        datafiles : list[Path]
            List of grib input filenames
        geo_coords: Callable[[UUID], dict[str, xr.DataArray]] | None
            Callable that returns the horizontal coordinates
            of the grid defined by the given UUID. The dimension must be "cell".
        ref_param : str
            name of parameter used to construct a reference grid

        Raises
        ------
        ValueError
            if the grid can not be constructed from the ref_param

        """
        return cls(
            data_source.FileDataSource(datafiles=[str(p) for p in datafiles]),
            geo_coords,
            ref_param,
        )

    def load(
        self,
        requests: Mapping[str, Request],
        extract_pv: str | None = None,
    ) -> dict[str, xr.DataArray]:
        """Load a dataset with the requested parameters.

        Parameters
        ----------
        requests : Mapping[str, Request]
            Mapping of label to request for a given field from the data source.
        extract_pv: str | None
            Optionally extract hybrid level coefficients from the field referenced by
            the given label.

        Raises
        ------
        RuntimeError
            if not all fields are found in the data source.

        Returns
        -------
        dict[str, xr.DataArray]
            Mapping of fields by label

        """
        result = {
            name: tasking.delayed(load_single_param)(
                self.data_source, req, self.geo_coords
            )
            for name, req in requests.items()
        }

        if extract_pv is not None:
            if extract_pv not in requests:
                msg = f"{extract_pv=} was not a key of the given requests."
                raise RuntimeError(msg)
            return result | metadata.extract_pv(result[extract_pv].metadata)
        return result

    def load_fieldnames(
        self,
        params: list[str],
        extract_pv: str | None = None,
    ) -> dict[str, xr.DataArray]:
        reqs = {param: param for param in params}
        return self.load(reqs, extract_pv)


def save(
    field: xr.DataArray,
    file_handle: io.BufferedWriter | io.BytesIO,
    bits_per_value: int = 16,
):
    """Write field to file in GRIB format.

    Parameters
    ----------
    field : xarray.DataArray
        Field to write into the output file.
    file_handle : io.BufferedWriter
        File handle for the output file.
    bits_per_value : int, optional
        Bits per value encoded in the output file. (Default: 16)

    Raises
    ------
    ValueError
        If the field does not have a metadata attribute.

    """
    if not hasattr(field, "metadata"):
        msg = "The metadata attribute is required to write to the GRIB format."
        raise ValueError(msg)

    md = field.metadata

    idx = {
        dim: field.coords[key]
        for key in field.dims
        if (dim := str(key)) not in {"x", "y"}
    }

    step_unit = UnitOfTime.MINUTE
    time_range_unit = UnitOfTime(md.get("indicatorOfUnitForTimeRange", 255)).unit
    time_range = _to_timedelta(md.get("lengthOfTimeRange", 0), unit=time_range_unit)

    if md.get("numberOfTimeRange", 1) != 1:
        raise NotImplementedError("Unsupported value for numberOfTimeRange")

    def to_grib(loc: dict[str, xr.DataArray]):
        grib_loc = {
            DIM_MAP[key]: value.item()
            for key, value in loc.items()
            if key not in {"ref_time", "lead_time"}
        }
        step_end = np.timedelta64(loc["lead_time"].item(), "ns")
        step_begin = step_end - time_range
        return grib_loc | {
            "indicatorOfUnitOfTimeRange": step_unit.value,
            "forecastTime": step_begin / _to_timedelta(1, step_unit.unit),
            "dataDate": loc["ref_time"].dt.strftime("%Y%m%d").item(),
            "dataTime": loc["ref_time"].dt.strftime("%H%M").item(),
        }

    for idx_slice in product(*idx.values()):
        loc = {dim: value for dim, value in zip(idx.keys(), idx_slice)}
        array = field.sel(loc).values
        metadata = md.override(to_grib(loc))

        fs = ekd.FieldList.from_numpy(array, metadata)
        fs.write(file_handle, bits_per_value=bits_per_value)


def get_code_flag(value: int, indices: Sequence[int]) -> list[bool]:
    """Get the code flag value at the given indices.

    Parameters
    ----------
    value : int
        The code flag as an integer in the [0, 255] range.
    indices : Sequence[int]
        Indices at which to get the flag values. Left to right, 1-based.

    Returns
    -------
    list[bool]
        The code flag values at the given indices.

    """
    if not 0 <= value <= 255:
        raise ValueError("value must be a single byte integer")

    result = []
    for index in indices:
        if not 1 <= index <= 8:
            raise ValueError("index must in range [1,8]")
        shift = 8 - index
        result.append(bool(value >> shift & 1))
    return result


def set_code_flag(indices: Sequence[int]) -> int:
    """Create code flag by setting bits at the given indices.

    Parameters
    ----------
    indices : Sequence[int]
        Indices at which to set the flag values. Left to right, 1-based.

    Returns
    -------
    int
        Code flag with bits set at the given indices.

    """
    value = 0
    for index in indices:
        if not 1 <= index <= 8:
            raise ValueError("index must in range [1,8]")
        shift = 8 - index
        value |= 1 << shift
    return value



"""Algorithm for computation of height of zero degree isotherm."""

# Standard library
from typing import cast

# Third-party
import numpy as np
import xarray as xr

# Local
from .. import metadata
from .destagger import destagger


def fhzerocl(
    t: xr.DataArray, hhl: xr.DataArray, extrapolate: bool = False
) -> xr.DataArray:
    """Height of the zero deg C isotherm in m amsl.

    The algorithm searches from the top of the atmosphere downwards.
    When extrapolation is enabled, the search may extend past the lowest
    layers of the model by means of a linear model. The resulting height
    may be below the surface of the earth.

    Parameters
    ----------
    t : xr.DataArray
        Air temperature in K.
    hhl : xr.DataArray
        Heights of the interfaces between vertical layers in m amsl.
    extrapolate : bool, optional
        Allow the extrapolation of the search below the lowest model layer.
        Defaults to False.

    Returns
    -------
    xr.DataArray
        Height of the zero deg C isotherm in m amsl.

    """
    # Physical constants
    t0 = 273.15

    # Heights of layer mid surfaces (where t is defined)
    hfl = destagger(hhl, "z")

    tkm1 = t.shift(z=1)

    # 3d field with values of height for those levels where temperature
    # is > 0 and it was < 0 on the level below. Otherwise values are NaN.
    height0 = hfl.where((t >= t0) & (tkm1 < t0))

    # The previous condition can be satisfied on multiple levels.
    # Take the k indices of the maximum height value where the condition is satisfied
    maxind = cast(xr.DataArray, height0.fillna(-1).argmax(dim="z"))

    if extrapolate:
        # The full column is below freezing
        below_ground = (t < t0).all("z")
        # Temperature is increasing with decreasing altitude
        positive_dt = t[{"z": -1}] - t[{"z": -2}] > 1e-10
        # Allow t0 to be outside of the [height1, height2] range
        cond = np.logical_and(below_ground, positive_dt)
        maxind = maxind.where(~cond, -1)

    # Compute the 2D fields with height values where T is > 0 and < 0 on level below
    height2 = hfl[{"z": maxind}]
    # Compute the 2D fields with height values where T is < 0 and > 0 on level above
    height1 = hfl[{"z": maxind - 1}]
    # The height level where T == t0 must be between [height1, height2]
    t2 = t[{"z": maxind}]
    t1 = tkm1[{"z": maxind}]

    hzerocl = height1 + (height2 - height1) * (t0 - t1) / (t2 - t1)

    return xr.DataArray(
        data=hzerocl.where(hzerocl > 0),
        attrs=metadata.override(t.metadata, shortName="HZEROCL"),
    )



"""algorithm for computing omega_slope."""

# Third-party
import numpy as np
import xarray as xr

# Local
from .. import metadata
from .. import physical_constants as pc


# similar to the subtract.accumulate but permute the order of the operans of the diff
# TODO implement as a ufunc
def cumdiff(A, axis):
    r = np.empty(np.shape(A))
    t = 0  # op = the ufunc being applied to A's  elements
    for i in range(np.shape(A)[axis]):
        t = np.take(A, i, axis) - t

        slices = []
        for dim in range(A.ndim):
            if dim == axis:
                slices.append(slice(i, i + 1))
            else:
                slices.append(slice(None))

        r[tuple(slices)] = np.expand_dims(t, axis=axis)
    return r


def omega_slope(
    ps: xr.DataArray, etadot: xr.DataArray, ak: xr.DataArray, bk: xr.DataArray
):
    """Compute the omega slope.

    Converts ECMWF etadot (deta/dt) to etadot * dp/deta, required by FLEXPART.

    Parameters
    ----------
    ps : xarray.DataArray
        Pressure (S) (not reduced) in Pa.
    etadot : xarray.DataArray
        Eta-coordinate vertical velocity (deta/dt) in s**-1.
    ak : xarray.DataArray
        Hybrid level A coefficient.
    bk : xarray.DataArray
        Hybrid level B coefficient.


    Returns
    -------
    xarray.DataArray
        Vertical velocity (pressure) in Pa s**-1.

    Notes
    -----
    Fieldextra returns the parameter as ETADOT.

    """
    dak_dz = ak.diff(dim="z")
    dbk_dz = bk.diff(dim="z")

    res = (
        2.0
        * etadot
        * ps
        * (dak_dz / ps + dbk_dz)
        / (dak_dz / pc.surface_pressure_ref + dbk_dz)
    ).reduce(cumdiff, dim="z")

    return xr.DataArray(
        data=res,
        attrs=metadata.override(
            # Eta-coordinate vertical velocity
            etadot.metadata,
            discipline=0,
            parameterCategory=2,
            parameterNumber=32,
        ),
    )
